from __future__ import annotations

import json
from datetime import datetime
from pathlib import Path
from typing import IO, Any, Optional

import requests
import requests.adapters
import urllib3
from pydantic import AnyUrl

from checker.exceptions import PluginExecutionFailed

from .base import PluginABC, PluginOutput


class ManytaskPlugin(PluginABC):
    """Given score report it to the manytask.
    Datetime format in args should be: '%Y-%m-%dT%H:%M:%S.%f%z'"""

    DEFAULT_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"

    name = "report_score_manytask"

    class Args(PluginABC.Args):
        origin: Optional[str] = None  # as pydantic does not support | in older python versions
        patterns: list[str] = ["*"]
        username: str
        task_name: str
        score: float  # TODO: validate score is in [0, 1] (bonus score is higher than 1)
        report_url: AnyUrl
        report_token: str
        check_deadline: bool
        send_time: datetime = datetime.now().astimezone()

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        output: list[str] = []

        if not args.send_time.tzinfo:
            output.append("Warning: No timezone provided for send_time, possible time miscalculations")
        send_time_formatted = args.send_time.strftime(self.DEFAULT_TIME_FORMAT)

        # Do not expose token in logs.
        data = {
            "token": args.report_token,
            "task": args.task_name,
            "username": args.username,
            "score": args.score,
            "check_deadline": args.check_deadline,
            "submit_time": send_time_formatted,
        }

        files = None
        if args.origin is not None:
            files = self._collect_files_to_send(args.origin, args.patterns)

        if verbose:
            output.append(str(files))

        response = self._post_with_retries(args.report_url, data, files)

        try:
            result = response.json()
            output.append(
                f"Report for task '{args.task_name}' for user '{args.username}', "
                f"requested score: {args.score}, result score: {result['score']}"
            )
            return PluginOutput(output="\n".join(output))
        except (json.JSONDecodeError, KeyError):
            raise PluginExecutionFailed("Unable to decode response")

    @staticmethod
    def _post_with_retries(
        report_url: AnyUrl,
        data: dict[str, Any],
        files: dict[str, tuple[str, IO[bytes]]] | None,
    ) -> requests.Response:
        retry_strategy = urllib3.Retry(total=3, backoff_factor=1, status_forcelist=[408, 500, 502, 503, 504])
        adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)
        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        response = session.post(url=f"{report_url}api/report", data=data, files=files)

        if response.status_code >= 400:
            raise PluginExecutionFailed(f"{response.status_code}: {response.text}")

        return response

    @staticmethod
    def _collect_files_to_send(origin: str, patterns: list[str]) -> dict[str, tuple[str, IO[bytes]]]:
        source_dir = Path(origin)
        return {
            path.name: (str(path.relative_to(source_dir)), open(path, "rb"))
            for pattern in patterns
            for path in source_dir.glob(pattern)
            if path.is_file()
        }



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel, ValidationError

from checker.exceptions import BadConfig, BadStructure


@dataclass
class PluginOutput:
    """Plugin output dataclass.
    :ivar output: str plugin output
    :ivar percentage: float plugin percentage
    """

    output: str
    percentage: float = 1.0


class PluginABC(ABC):
    """Abstract base class for plugins.
    :ivar name: str plugin name, searchable by this name
    """

    name: str

    class Args(BaseModel):
        """Base class for plugin arguments.
        You have to subclass this class in your plugin.
        """

        pass

    def run(self, args: dict[str, Any], *, verbose: bool = False) -> PluginOutput:
        """Run the plugin.
        :param args: dict plugin arguments to pass to subclass Args
        :param verbose: if True should print teachers debug info, if False student mode
        :raises BadConfig: if plugin arguments are invalid
        :raises ExecutionFailedError: if plugin failed
        :return: PluginOutput with stdout/stderr and percentage
        """
        args_obj = self.Args(**args)

        return self._run(args_obj, verbose=verbose)

    @classmethod
    def validate(cls, args: dict[str, Any]) -> None:
        """Validate the plugin arguments.
        :param args: dict plugin arguments to pass to subclass Args
        :raises BadConfig: if plugin arguments are invalid
        :raises BadStructure: if _run method is not implemented
        """
        try:
            cls.Args(**args)
        except ValidationError as e:
            raise BadConfig(f"Plugin {cls.name} arguments validation error:\n{e}")

        if not hasattr(cls, "_run"):
            raise BadStructure(f"Plugin {cls.name} does not implement _run method")

    @abstractmethod
    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:
        """Actual run the plugin.
        You have to implement this method in your plugin.
        In case of failure, raise ExecutionFailedError with an error message and output.
        :param args: plugin arguments, see Args subclass
        :param verbose: if True should print teachers debug info, if False student mode
        :return: PluginOutput with stdout/stderr and percentage
        :raises ExecutionFailedError: if plugin failed
        """
        pass



from __future__ import annotations

import sys
from inspect import cleandoc
from typing import Any


def print_ascii_tag(
    version: str | None = None,
    file: Any = None,
    color: str = "pink",
    **kwargs: Any,
) -> None:
    # https://patorjk.com/software/taag/#p=display&f=Doom&t=%20%20%20%20Manytask%20%20Checker
    ascii_tag = cleandoc(
        r"""
    
    ___  ___                  _            _       _____ _               _             
    |  \/  |                 | |          | |     /  __ \ |             | |            
    | .  . | __ _ _ __  _   _| |_ __ _ ___| | __  | /  \/ |__   ___  ___| | _____ _ __ 
    | |\/| |/ _` | '_ \| | | | __/ _` / __| |/ /  | |   | '_ \ / _ \/ __| |/ / _ \ '__|
    | |  | | (_| | | | | |_| | || (_| \__ \   <   | \__/\ | | |  __/ (__|   <  __/ |   
    \_|  |_/\__,_|_| |_|\__, |\__\__,_|___/_|\_\   \____/_| |_|\___|\___|_|\_\___|_|   
                         __/ |                                                         
                        |___/                                                          
    
    """
    )
    print_info(ascii_tag, color=color, file=file, **kwargs)
    if version:
        print_info(f"{version}", color=color, file=file, **kwargs)


def print_info(
    *args: Any,
    file: Any = None,
    color: str | None = None,
    **kwargs: Any,
) -> None:
    colors = {
        "white": "\033[97m",
        "cyan": "\033[96m",
        "pink": "\033[95m",
        "blue": "\033[94m",
        "orange": "\033[93m",
        "green": "\033[92m",
        "red": "\033[91m",
        "grey": "\033[90m",
        "endc": "\033[0m",
    }

    file = file or sys.stderr

    data = " ".join(map(str, args))
    if color in colors:
        print(colors[color] + data + colors["endc"], file=file, **kwargs)
    else:
        print(data, file=file, **kwargs)
    file.flush()


def print_separator(
    symbol: str,
    file: Any = None,
    color: str = "pink",
    string_length: int = 80,
) -> None:
    print_info(symbol * string_length, color=color)


def print_header_info(
    header_string: str,
    file: Any = None,
    color: str = "pink",
    string_length: int = 80,
    **kwargs: Any,
) -> None:
    info_extended_string = " " + header_string + " "
    print_info("", file=file)
    print_separator(symbol="+", string_length=string_length, color=color, file=file)
    print_info(f"{info_extended_string:+^{string_length}}", color=color, file=file)
    print_separator(symbol="+", string_length=string_length, color=color, file=file)



from __future__ import annotations

import importlib
import importlib.util
import pkgutil
import sys
from collections.abc import Sequence
from pathlib import Path

from checker.utils import print_info

from .base import PluginABC, PluginOutput  # noqa: F401


__all__ = [
    "PluginABC",
    "PluginOutput",
    "load_plugins",
]


def get_all_subclasses(cls: type[PluginABC]) -> set[type[PluginABC]]:
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in get_all_subclasses(c)])


def load_plugins(
    search_directories: Sequence[str | Path] | None = None,
    *,
    verbose: bool = False,
) -> dict[str, type[PluginABC]]:
    """
    Load plugins from the plugins directory.
    :param search_directories: list of directories to search for plugins
    :param verbose: verbose output
    """
    search_directories = search_directories or []
    search_directories = [
        Path(__file__).parent,
        *search_directories,
    ]  # add local plugins first

    # force load plugins
    print_info("Loading plugins...")
    for module_info in pkgutil.iter_modules([str(path) for path in search_directories]):
        if module_info.name == "__init__":
            continue
        if verbose:
            print_info(f"- {module_info.name} from {module_info.module_finder.path}")  # type: ignore[union-attr]

        spec = module_info.module_finder.find_spec(fullname=module_info.name)  # type: ignore[call-arg]
        if spec is None:
            raise ImportError(f"Could not find {module_info.name}")
        module = importlib.util.module_from_spec(spec)
        module.__package__ = __package__  # TODO: check for external plugins

        sys.modules[module_info.name] = module
        assert spec.loader is not None
        spec.loader.exec_module(module)

    # collect plugins as abstract class subclasses
    plugins = {}
    for subclass in get_all_subclasses(PluginABC):  # type: ignore[type-abstract]
        plugins[subclass.name] = subclass
    if verbose:
        print_info(f"Loaded: {', '.join(plugins.keys())}")
    return plugins



from __future__ import annotations

import os
from pathlib import Path
from typing import Union

from ..exceptions import PluginExecutionFailed
from .base import PluginOutput
from .scripts import PluginABC, RunScriptPlugin


HOME_PATH = str(Path.home())


class SafeRunScriptPlugin(PluginABC):
    """Wrapper over RunScriptPlugin to run students scripts safety.
    Plugin uses Firejail tool to create sandbox for the running process.
    It allows hide environment variables and control access to network and file system.
    If `allow_fallback=True` then if Firejail is not installed, it will fallback to RunScriptPlugin.
    """

    name = "safe_run_script"

    class Args(PluginABC.Args):
        origin: str
        script: Union[str, list[str]]  # as pydantic does not support | in older python versions
        timeout: Union[float, None] = None  # as pydantic does not support | in older python versions

        env_whitelist: list[str] = list()
        paths_whitelist: list[str] = list()
        lock_network: bool = True
        allow_fallback: bool = False

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        import subprocess

        # test if firejail script is available
        # TODO: test fallback
        result = subprocess.run(["firejail", "--version"], capture_output=True)
        if result.returncode != 0:
            if args.allow_fallback:
                # fallback to RunScriptPlugin
                run_args = RunScriptPlugin.Args(
                    origin=args.origin, script=args.script, timeout=args.timeout, env_whitelist=args.env_whitelist
                )
                output = RunScriptPlugin()._run(args=run_args, verbose=verbose)
                if verbose:
                    output.output = f"Firejail is not installed. Fallback to RunScriptPlugin.\n{output.output}"
                return output
            else:
                # error
                raise PluginExecutionFailed("Firejail is not installed", output=result.stderr.decode("utf-8"))

        # Construct firejail command
        command: list[str] = ["firejail", "--quiet", "--noprofile"]

        # lock network access
        if args.lock_network:
            command.append("--net=none")

        # Collect all allow paths
        allow_paths = {*args.paths_whitelist, args.origin}
        # a bit tricky but if paths is only /tmp add ~/tmp instead of it
        if "/tmp" in allow_paths and len(allow_paths) == 1:
            allow_paths.add("~/tmp")
        # remove /tmp from paths as it causes error inside Firejail
        if "/tmp" in allow_paths:
            allow_paths.remove("/tmp")
        # replace ~ by the full home path
        for path in allow_paths:
            full_path = path if not path.startswith("~") else HOME_PATH + path[1:]
            # allow access to origin dir
            command.append(f"--whitelist={full_path}")

        # Hide all environment variables except allowed
        command.append("env -i")
        for env in args.env_whitelist:
            command.append(f'{env}="{os.environ.get(env, "")}"')

        # create actual command
        run_command: str | list[str]
        if isinstance(args.script, str):
            run_command = " ".join(command) + " " + args.script
        elif isinstance(args.script, list):
            run_command = command + args.script
        else:
            assert False, "Now Reachable"

        # Will use RunScriptPlugin to run Firejail+command
        run_args = RunScriptPlugin.Args(
            origin=args.origin, script=run_command, timeout=args.timeout, env_whitelist=None
        )
        return RunScriptPlugin()._run(args=run_args, verbose=verbose)



from __future__ import annotations

import sys
from datetime import datetime, timedelta
from enum import Enum
from typing import Optional, Union


if sys.version_info < (3, 8):
    from pytz import ZoneInfoNotFoundError as ZoneInfoNotFoundError, timezone as ZoneInfo
else:
    from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from pydantic import AnyUrl, Field, field_validator, model_validator

from .utils import CustomBaseModel, YamlLoaderMixin


class ManytaskSettingsConfig(CustomBaseModel):
    """Manytask settings."""

    course_name: str
    gitlab_base_url: AnyUrl
    public_repo: str
    students_group: str


class ManytaskUiConfig(CustomBaseModel):
    task_url_template: str  # $GROUP_NAME $TASK_NAME vars are available
    links: dict[str, str] = Field(default_factory=dict)

    @field_validator("task_url_template")
    @classmethod
    def check_task_url_template(cls, data: str | None) -> str | None:
        if data is not None and (not data.startswith("http://") and not data.startswith("https://")):
            raise ValueError("task_url_template should be http or https")
        # if data is not None and "$GROUP_NAME" not in data and "$TASK_NAME" not in data:
        #     raise ValueError("task_url should contain at least one of $GROUP_NAME and $TASK_NAME vars")
        return data


class ManytaskDeadlinesType(Enum):
    HARD = "hard"
    INTERPOLATE = "interpolate"


class ManytaskTaskConfig(CustomBaseModel):
    task: str

    enabled: bool = True

    score: int
    special: int = 0

    is_bonus: bool = False
    is_special: bool = False

    # Note: use Optional/Union[...] instead of ... | None as pydantic does not support | in older python versions
    url: Optional[AnyUrl] = None

    @property
    def name(self) -> str:
        return self.task


class ManytaskGroupConfig(CustomBaseModel):
    group: str

    enabled: bool = True

    # Note: use Optional/Union[...] instead of ... | None as pydantic does not support | in older python versions
    start: datetime
    steps: dict[float, Union[datetime, timedelta]] = Field(default_factory=dict)
    end: Union[datetime, timedelta]

    tasks: list[ManytaskTaskConfig] = Field(default_factory=list)

    @property
    def name(self) -> str:
        return self.group

    def get_percents_before_deadline(self) -> dict[float, datetime]:
        return {
            percent: date_or_delta if isinstance(date_or_delta, datetime) else self.start + date_or_delta
            for percent, date_or_delta in zip([1.0, *self.steps.keys()], [*self.steps.values(), self.end])
        }

    def get_current_percent_multiplier(self, now: datetime) -> float:
        percents = self.get_percents_before_deadline()
        for percent, date in percents.items():
            if now <= date:
                return percent
        return 0.0

    def replace_timezone(self, timezone: ZoneInfo) -> None:
        self.start = self.start.replace(tzinfo=timezone)
        self.end = self.end.replace(tzinfo=timezone) if isinstance(self.end, datetime) else self.end
        self.steps = {k: v.replace(tzinfo=timezone) for k, v in self.steps.items() if isinstance(v, datetime)}

    @model_validator(mode="after")
    def check_dates(self) -> "ManytaskGroupConfig":
        # check end
        if isinstance(self.end, timedelta) and self.end < timedelta():
            raise ValueError(f"end timedelta <{self.end}> should be positive")
        if isinstance(self.end, datetime) and self.end < self.start:
            raise ValueError(f"end datetime <{self.end}> should be after the start <{self.start}>")

        # check steps
        last_step_date_or_delta: datetime | timedelta = self.start
        for _, date_or_delta in self.steps.items():
            step_date = self.start + date_or_delta if isinstance(date_or_delta, timedelta) else date_or_delta
            last_step_date = (
                self.start + last_step_date_or_delta
                if isinstance(last_step_date_or_delta, timedelta)
                else last_step_date_or_delta
            )

            if isinstance(date_or_delta, timedelta) and date_or_delta < timedelta():
                raise ValueError(f"step timedelta <{date_or_delta}> should be positive")
            if isinstance(date_or_delta, datetime) and date_or_delta <= self.start:
                raise ValueError(f"step datetime <{date_or_delta}> should be after the start {self.start}")

            if step_date <= last_step_date:
                raise ValueError(
                    f"step datetime/timedelta <{date_or_delta}> "
                    f"should be after the last step <{last_step_date_or_delta}>"
                )
            last_step_date_or_delta = date_or_delta

        return self


class ManytaskDeadlinesConfig(CustomBaseModel):
    timezone: str

    # Note: use Optional/Union[...] instead of ... | None as pydantic does not support | in older python versions
    deadlines: ManytaskDeadlinesType = ManytaskDeadlinesType.HARD
    max_submissions: Optional[int] = None
    submission_penalty: float = 0

    schedule: list[ManytaskGroupConfig]  # list of groups with tasks

    def get_now_with_timezone(self) -> datetime:
        return datetime.now(tz=ZoneInfo(self.timezone))

    @field_validator("max_submissions")
    @classmethod
    def check_max_submissions(cls, data: int | None) -> int | None:
        if data is not None and data <= 0:
            raise ValueError("max_submissions should be positive")
        return data

    @field_validator("submission_penalty")
    @classmethod
    def check_submission_penalty(cls, data: float) -> float:
        if data < 0:
            raise ValueError("submission_penalty should be non-negative")
        return data

    @field_validator("timezone")
    @classmethod
    def check_valid_timezone(cls, timezone: str) -> str:
        try:
            ZoneInfo(timezone)
        except ZoneInfoNotFoundError as e:
            raise ValueError(str(e))
        return timezone

    @field_validator("schedule")
    @classmethod
    def check_group_task_names_unique(cls, data: list[ManytaskGroupConfig]) -> list[ManytaskGroupConfig]:
        group_names = [group.name for group in data]
        tasks_names = [task.name for group in data for task in group.tasks]

        # group names unique
        group_names_duplicates = [name for name in group_names if group_names.count(name) > 1]
        if group_names_duplicates:
            raise ValueError(f"Group names should be unique, duplicates: {group_names_duplicates}")

        # task names unique
        tasks_names_duplicates = [name for name in tasks_names if tasks_names.count(name) > 1]
        if tasks_names_duplicates:
            raise ValueError(f"Task names should be unique, duplicates: {tasks_names_duplicates}")

        # # group names and task names not intersect (except single task in a group with the same name)
        # no_single_task_groups = [group for group in data if not (len(group.tasks) == 1
        # and group.name == group.tasks[0].name)]

        return data

    @model_validator(mode="after")
    def set_timezone(self) -> "ManytaskDeadlinesConfig":
        timezone = ZoneInfo(self.timezone)
        for group in self.schedule:
            group.replace_timezone(timezone)
        return self

    def find_task(self, task_name: str) -> tuple[ManytaskGroupConfig, ManytaskTaskConfig]:
        for group in self.schedule:
            for task in group.tasks:
                if task.name == task_name:
                    return group, task
        raise KeyError(f"Task {task_name} not found")

    def get_groups(
        self,
        enabled: bool | None = None,
        started: bool | None = None,
        *,
        now: datetime | None = None,
    ) -> list[ManytaskGroupConfig]:
        if now is None:
            now = self.get_now_with_timezone()

        groups = [group for group in self.schedule]

        if enabled is not None:
            groups = [group for group in groups if group.enabled == enabled]

        if started is not None:
            if started:
                groups = [group for group in groups if group.start <= now]
            else:
                groups = [group for group in groups if group.start > now]

        return groups

    def get_tasks(
        self,
        enabled: bool | None = None,
        started: bool | None = None,
        is_bonus: bool | None = None,
        *,
        now: datetime | None = None,
    ) -> list[ManytaskTaskConfig]:
        # TODO: refactor
        if now is None:
            now = self.get_now_with_timezone()

        groups = self.get_groups(started=started, now=now)

        if enabled is True:
            groups = [group for group in groups if group.enabled]
            extra_tasks = []
        elif enabled is False:
            groups = groups
            extra_tasks = [task for group in groups for task in group.tasks if not group.enabled]
        else:  # None
            groups = groups
            extra_tasks = []

        tasks = [task for group in groups for task in group.tasks]

        if enabled is not None:
            tasks = [task for task in tasks if task.enabled == enabled]

        for extra_task in extra_tasks:
            if extra_task not in tasks:
                tasks.append(extra_task)

        if is_bonus is not None:
            tasks = [task for task in tasks if task.is_bonus == is_bonus]

        return tasks

    def max_score(self, started: bool | None = True, *, now: datetime | None = None) -> int:
        return sum(task.score for task in self.get_tasks(enabled=True, started=started, is_bonus=False, now=now))

    @property
    def max_score_started(self) -> int:
        return self.max_score(started=True, now=self.get_now_with_timezone())


class ManytaskConfig(CustomBaseModel, YamlLoaderMixin["ManytaskConfig"]):
    """Manytask configuration."""

    version: int  # if config exists, version is always present

    settings: ManytaskSettingsConfig
    ui: ManytaskUiConfig
    deadlines: ManytaskDeadlinesConfig

    def get_groups(
        self,
        enabled: bool | None = None,
        started: bool | None = None,
        *,
        now: datetime | None = None,
    ) -> list[ManytaskGroupConfig]:
        return self.deadlines.get_groups(enabled=enabled, started=started, now=now)

    def get_tasks(
        self,
        enabled: bool | None = None,
        started: bool | None = None,
        *,
        now: datetime | None = None,
    ) -> list[ManytaskTaskConfig]:
        return self.deadlines.get_tasks(enabled=enabled, started=started, now=now)

    @field_validator("version")
    @classmethod
    def check_version(cls, data: int) -> int:
        if data != 1:
            raise ValueError(f"Only version 1 is supported for {cls.__name__}")
        return data



from .conftest import T_GENERATE_FILE_STRUCTURE


class TestGenerateFileStructure:
    def test_simple_case(self, generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> None:
        files_content = {
            "test.py": "print('Hello')\n",
            "test.txt": "Hello\n",
        }
        root = generate_file_structure(files_content)
        assert root.is_dir()
        assert (root / "test.py").is_file()
        assert (root / "test.txt").is_file()
        assert (root / "test.py").read_text() == "print('Hello')\n"
        assert (root / "test.txt").read_text() == "Hello\n"

    def test_subfolders(self, generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> None:
        files_content = {
            "folder": {
                "test.py": "print('Hello')\n",
                "test.txt": "Hello\n",
            },
            "test.txt": "Hello\n",
            "deep_folder": {
                "folder": {
                    "test.py": "print('Hello')\n",
                    "test.txt": "Hello\n",
                },
                "test.txt": "Hello\n",
            },
        }
        root = generate_file_structure(files_content)
        assert root.is_dir()
        assert (root / "folder").is_dir()
        assert (root / "test.txt").is_file()
        assert (root / "folder" / "test.py").is_file()
        assert (root / "folder" / "test.txt").is_file()
        assert (root / "deep_folder").is_dir()
        assert (root / "deep_folder" / "folder").is_dir()
        assert (root / "deep_folder" / "test.txt").is_file()
        assert (root / "deep_folder" / "folder" / "test.py").is_file()
        assert (root / "deep_folder" / "folder" / "test.txt").is_file()

    def test_empty_files(self, generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> None:
        files_content = {
            "test.py": "",
            "test.txt": "",
        }
        root = generate_file_structure(files_content)
        assert root.is_dir()
        assert (root / "test.py").is_file()
        assert (root / "test.txt").is_file()
        assert (root / "test.py").read_text() == ""
        assert (root / "test.txt").read_text() == ""

    def test_empty_directory(self, generate_file_structure: T_GENERATE_FILE_STRUCTURE) -> None:
        files_content = {
            "folder": {},
            "deep_folder": {
                "folder": {},
            },
            "test.txt": "",
        }
        root = generate_file_structure(files_content)
        assert root.is_dir()
        assert (root / "folder").is_dir()
        assert not list((root / "folder").iterdir())
        assert (root / "test.txt").is_file()
        assert (root / "deep_folder").is_dir()
        assert (root / "deep_folder" / "folder").is_dir()
        assert not list((root / "deep_folder" / "folder").iterdir())



from __future__ import annotations

import warnings
from collections.abc import Generator
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any

import git

from .configs import CheckerSubConfig, CheckerTestingConfig, ManytaskConfig
from .exceptions import BadConfig, CheckerException
from .utils import print_info


@dataclass
class FileSystemTask:
    name: str
    relative_path: str
    config: CheckerSubConfig


@dataclass
class FileSystemGroup:
    name: str
    relative_path: str
    config: CheckerSubConfig
    tasks: list[FileSystemTask]


class Course:
    """
    Class operates deadlines (filter, search etc), timezones and mapping tasks and groups to file system.
    Only operates with tasks and groups existing in file system.
    """

    TASK_CONFIG_NAME = ".task.yml"
    GROUP_CONFIG_NAME = ".group.yml"

    def __init__(
        self,
        manytask_config: ManytaskConfig,
        repository_root: Path,
        reference_root: Path | None = None,
        branch_name: str | None = None,
    ):
        self.manytask_config = manytask_config

        self.repository_root = repository_root
        self.reference_root = reference_root or repository_root

        self.potential_groups = {group.name: group for group in self._search_for_groups_by_configs(self.reference_root)}
        self.potential_tasks = {task.name: task for task in self._search_for_tasks_by_configs(self.reference_root)}

        self.branch_name = branch_name

    def validate(self) -> None:
        # check all groups and tasks mentioned in deadlines exists
        deadlines_groups = self.manytask_config.get_groups(enabled=True)
        for deadline_group in deadlines_groups:
            if deadline_group.name not in self.potential_groups:
                warnings.warn(f"Group {deadline_group.name} not found in repository")

        deadlines_tasks = self.manytask_config.get_tasks(enabled=True)
        for deadlines_task in deadlines_tasks:
            if deadlines_task.name not in self.potential_tasks:
                raise BadConfig(f"Task {deadlines_task.name} not found in repository")

    def get_groups(
        self,
        enabled: bool | None = None,
        started: bool | None = None,
        *,
        now: datetime | None = None,
    ) -> list[FileSystemGroup]:
        search_deadlines_groups = self.manytask_config.get_groups(enabled=enabled, started=started, now=now)

        return [
            self.potential_groups[deadline_group.name]
            for deadline_group in search_deadlines_groups
            if deadline_group.name in self.potential_groups
        ]

    def get_tasks(
        self,
        enabled: bool | None = None,
        started: bool | None = None,
        *,
        now: datetime | None = None,
    ) -> list[FileSystemTask]:
        search_deadlines_tasks = self.manytask_config.get_tasks(enabled=enabled, started=started, now=now)

        return [
            self.potential_tasks[deadline_task.name]
            for deadline_task in search_deadlines_tasks
            if deadline_task.name in self.potential_tasks
        ]

    @staticmethod
    def _search_for_tasks_by_configs(
        root: Path,
    ) -> Generator[FileSystemTask, Any, None]:
        for task_config_path in root.glob(f"**/{Course.TASK_CONFIG_NAME}"):
            relative_task_path = task_config_path.parent.relative_to(root)

            # if empty file - use default
            if task_config_path.read_text().strip() == "":
                task_config = CheckerSubConfig.default()
            # if any content - read yml
            else:
                task_config = CheckerSubConfig.from_yaml(task_config_path)

            yield FileSystemTask(
                name=task_config_path.parent.name,
                relative_path=str(relative_task_path),
                config=task_config,
            )

    @staticmethod
    def _search_for_groups_by_configs(
        root: Path,
    ) -> Generator[FileSystemGroup, Any, None]:
        for group_config_path in root.glob(f"**/{Course.GROUP_CONFIG_NAME}"):
            relative_group_path = group_config_path.parent.relative_to(root)

            # if empty file - use default
            if group_config_path.read_text().strip() == "" or group_config_path.read_text().strip() == "\n":
                group_config = CheckerSubConfig.default()
            # if any content - read yml
            else:
                group_config = CheckerSubConfig.from_yaml(group_config_path)

            group_tasks = list(Course._search_for_tasks_by_configs(group_config_path.parent))
            for task in group_tasks:
                task.relative_path = str(relative_group_path / task.relative_path)

            yield FileSystemGroup(
                name=group_config_path.parent.name,
                relative_path=str(relative_group_path),
                config=group_config,
                tasks=group_tasks,
            )

    def detect_changes(
        self,
        detection_type: CheckerTestingConfig.ChangesDetectionType,
    ) -> list[FileSystemTask]:
        """
        Detects changes in the repository based on the provided detection type.

        :param detection_type: detection type, see CheckerTestingConfig.ChangesDetectionType
            - BRANCH_NAME: task name == branch name (single task/group)
            - COMMIT_MESSAGE: task name in commit message (can be multiple tasks/groups)
            - LAST_COMMIT_CHANGES: task relative path in last commit changes (can be multiple tasks)
        :return: list of changed tasks
        :raises CheckerException: if repository is not a git repository
        """
        print_info(f"Detecting changes by {detection_type}")
        potential_tasks = self.get_tasks(enabled=True)
        enabled_groups = self.manytask_config.get_groups(
            enabled=True
        )  # 'cause we want to check no-folder groups as well

        try:
            repo = git.Repo(self.repository_root)
        except git.exc.InvalidGitRepositoryError:
            raise CheckerException(f"Git Repository in {self.repository_root} not found")

        if detection_type == CheckerTestingConfig.ChangesDetectionType.BRANCH_NAME:
            try:
                branch_name = repo.active_branch.name
            except TypeError:
                if self.branch_name is None:
                    raise CheckerException("Detached HEAD state and no branch name provided")
                branch_name = self.branch_name
            print_info(f"Branch name: {branch_name}", color="grey")

            # try to get groups first
            changed_enabled_groups = [group for group in enabled_groups if group.name == branch_name]
            if changed_enabled_groups:
                print_info(
                    f"Changed groups: {[g.name for g in changed_enabled_groups]} (branch name == group name)",
                    color="grey",
                )
                changed_enabled_tasks_names = {task.name for group in changed_enabled_groups for task in group.tasks}
                changed_tasks = [task for task in potential_tasks if task.name in changed_enabled_tasks_names]
                print_info(
                    f"Changed tasks: {[t.name for t in changed_tasks]} (branch name == group name)",
                    color="grey",
                )
                return changed_tasks

            # if no groups found, try to get tasks
            changed_tasks = [task for task in potential_tasks if task.name == branch_name]
            print_info(
                f"Changed tasks: {[t.name for t in changed_tasks]} (branch name == task/group name)",
                color="grey",
            )
            if not changed_tasks:
                print_info(f"No active task/group found for branch {branch_name}", color="yellow")

            return changed_tasks

        elif detection_type == CheckerTestingConfig.ChangesDetectionType.COMMIT_MESSAGE:
            commit_message = repo.head.commit.message
            if isinstance(commit_message, bytes):  # pragma: no cover
                commit_message = commit_message.decode("utf-8")
            print_info(f"Commit message: {commit_message}", color="grey")

            # try to get groups first
            changed_enabled_groups = [group for group in enabled_groups if group.name in commit_message]
            if changed_enabled_groups:
                print_info(
                    f"Changed groups: {[g.name for g in changed_enabled_groups]} (group name in commit message)",
                    color="grey",
                )
                changed_enabled_tasks_names = {task.name for group in changed_enabled_groups for task in group.tasks}
                changed_tasks = [task for task in potential_tasks if task.name in changed_enabled_tasks_names]
                print_info(
                    f"Changed tasks: {[t.name for t in changed_tasks]} (group name in commit message)",
                    color="grey",
                )
                return changed_tasks

            # if no groups found, try to get tasks
            changed_tasks = [task for task in potential_tasks if task.name in commit_message]
            print_info(
                f"Changed tasks: {[t.name for t in changed_tasks]} (task name in commit message)",
                color="grey",
            )
            if not changed_tasks:
                print_info(f"No active tasks/groups found for commit message {commit_message}", color="yellow")

            return changed_tasks

        elif detection_type == CheckerTestingConfig.ChangesDetectionType.LAST_COMMIT_CHANGES:
            last_commit = repo.head.commit
            changed_files = [item.a_path for item in last_commit.diff("HEAD~1")]
            print_info(f"Last commit changes: {changed_files}", color="grey")

            changed_tasks = [
                task for task in potential_tasks if any(task.relative_path in file for file in changed_files)
            ]
            print_info(
                f"Changed tasks: {[t.name for t in changed_tasks]} (changed files in last commit)",
                color="grey",
            )
            if not changed_tasks:
                warnings.warn(f"No active tasks found for last commit changes {changed_files}")

            return changed_tasks

        else:  # pragma: no cover
            assert False, "Unreachable code"



# type: ignore
from __future__ import annotations

import os
import re
import subprocess
import time
from pathlib import Path
from tempfile import TemporaryDirectory

import gitlab.v4.objects

from ..course import CourseConfig
from ..utils.glab import GITLAB, GITLAB_HOST_URL, MASTER_BRANCH, get_private_project, get_public_project
from ..utils.print import print_info


MR_COPY_TOKEN = os.environ.get('MR_COPY_TOKEN')


def _student_mr_title_generator(merge_request: gitlab.v4.objects.MergeRequest) -> str:
    iid = merge_request.iid
    title = merge_request.title
    username = merge_request.author['username']

    return f'[STUDENT {username} MR {iid}] {title}'


def _get_student_mr_title_prefix(full_title: str) -> str:
    _title_search = re.match(r'^[.*]', full_title)
    assert _title_search
    prefix = _title_search.group(0)
    return prefix


def _student_mr_branch_name_generator(merge_request: gitlab.v4.objects.MergeRequest) -> str:
    iid = merge_request.iid
    username = merge_request.author['username']
    return f'students/{username}/mr-{iid}'


def _student_mr_desc_generator(merge_request: gitlab.v4.objects.MergeRequest) -> str:
    source_branch = merge_request.source_branch
    target_branch = merge_request.target_branch
    iid = merge_request.iid
    title = merge_request.title
    url = merge_request.web_url
    description = merge_request.description
    username = merge_request.author['username']
    name = merge_request.author['name']
    user_url = merge_request.author['web_url']
    return '  \n'.join([
        f'Contribution by student [{username}]({user_url}): {name}',
        '-'*16,
        f'Original [MR{iid}]({url}) from {source_branch} to {target_branch} in public project',
        f'Title: {title}',
        f'Branch: {source_branch}',
        '',
        f'{description}',
    ])


def _get_student_mr_source_url(merge_request: gitlab.v4.objects.MergeRequest) -> tuple[str, str]:
    source_project_id = merge_request.source_project_id
    full_project = GITLAB.projects.get(source_project_id)

    source_branch = merge_request.source_branch

    return ''.join([
        'https://',
        f'kblack:{MR_COPY_TOKEN}@',
        GITLAB_HOST_URL.removeprefix('https://'),
        '/',
        full_project.path_with_namespace,
    ]), source_branch


def copy_merge_requests(course_config: CourseConfig, dry_run: bool = False) -> None:
    """Copy changes from all open MR"""
    raise NotImplementedError

    private_project = get_private_project(course_config.private_group, course_config.private_repo)
    full_private_project = GITLAB.projects.get(private_project.id)
    public_project = get_public_project(course_config.private_group, course_config.public_repo)
    full_public_project = GITLAB.projects.get(public_project.id)

    # Run rebase
    print_info('Go and rebase public project', color='pink')
    public_mrs = full_public_project.mergerequests.list(state='opened')
    for mr in public_mrs:
        print_info(f'Rebase public MR {mr.iid} "{mr.title}":', color='pink')
        # mr.notes.create({'body': 'Run auto rebase...'})
        print_info(mr, color='grey')
        rebase_in_progress = mr.rebase()['rebase_in_progress']
        print_info("rebase_in_progress", rebase_in_progress)

    print_info('Waiting until rebase done...', color='pink')
    time.sleep(60)  # TODO: add check

    stdout = subprocess.run(
        'git fetch private',
        encoding='utf-8',
        shell=True, check=True,
        stdout=subprocess.PIPE,
    ).stdout
    print_info(stdout, color='grey')

    private_mrs = full_private_project.mergerequests.list(state='opened')
    private_students_mr = {
        prefix_match.group(0): mr for mr in private_mrs if (prefix_match := re.match(r'^\[.*\]', mr.title))
    }
    print_info('private_students_mr', len(private_students_mr), private_students_mr.keys())

    public_mrs = full_public_project.mergerequests.list(state='opened')

    print_info('Throughout public MR', color='pink')
    for mr in public_mrs:
        full_mr = GITLAB.mergerequests.get(mr.id)
        print_info('full_mr', full_mr)
        print_info(f'Look up public MR {mr.iid} "{mr.title}":', color='pink')
        print_info(mr, color='grey')

        gen_mr_full_title = _student_mr_title_generator(mr)
        gen_mr_prefix = _get_student_mr_title_prefix(gen_mr_full_title)

        if gen_mr_prefix in private_students_mr:
            print_info(f'MR "{gen_mr_prefix}" exists. Updating it...')
            # Already exists. Get branch name
            private_mr = private_students_mr[gen_mr_prefix]
            private_branch_name = private_mr.source_branch

            # An checkout existed branch
            print_info(f'checkout {private_branch_name}...')
            stdout = subprocess.run(
                f'git checkout --force --track private/{private_branch_name}',
                encoding='utf-8',
                shell=True, check=True,
                stdout=subprocess.PIPE,
            ).stdout
            print_info(stdout, color='grey')

            # Update desc and so on
            print_info('updating mr...')
            private_mr.title = _student_mr_title_generator(mr)
            private_mr.description = _student_mr_desc_generator(mr)
            private_mr.labels = ['contributing']
            private_mr.save()
            print_info('private_mr', private_mr, color='grey')
        else:
            # Not exist. So create
            print_info(f'MR "{gen_mr_prefix}" not exists. Creating it...')
            private_branch_name = _student_mr_branch_name_generator(mr)

            # Create and push branch
            print_info(f'create and checkout {private_branch_name}...')
            stdout = subprocess.run(
                f'git checkout --force -b {private_branch_name} private/{MASTER_BRANCH} '
                f'&& git push --set-upstream  private {private_branch_name}',
                encoding='utf-8',
                shell=True, check=True,
                stdout=subprocess.PIPE,
            ).stdout
            print_info(stdout, color='grey')

            # Create a MR
            print_info('creating mr...')
            private_mr = full_private_project.mergerequests.create({
                'source_branch': private_branch_name,
                'target_branch': MASTER_BRANCH,
                'title': _student_mr_title_generator(mr),
                'description': _student_mr_desc_generator(mr),
                'labels': ['contributing'],
                'remove_source_branch': True,
                'squash': True,
                'allow_maintainer_to_push': True,
            })
            print_info('private_mr', private_mr, color='grey')

        # Del processed private MR (leave only outdated)
        if gen_mr_prefix in private_students_mr:
            del private_students_mr[gen_mr_prefix]

        print_info(f'git status in {private_branch_name}:')
        stdout = subprocess.run(
            'git status',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info('Current status:', color='grey')
        print_info(stdout, color='grey')

        tmp_dir = TemporaryDirectory(dir=Path('.'))
        tmp_dir_path = tmp_dir.name
        # print_info(f'clone changes in {private_branch_name} from public mr:')
        # tmp_dir_path = mkdtemp(dir=Path('.'))

        source_url, source_branch = _get_student_mr_source_url(mr)
        # print(source_url, source_branch, tmp_dir_path)

        # Clone single branch in tmp folder
        # TODO: not copy, but merge
        print_info('Clone and update files...')
        stdout = subprocess.run(
            f'git clone --depth=1 --branch {source_branch} {source_url} {tmp_dir_path} && rm -rf {tmp_dir_path}/.git &&'
            f'cp -a {tmp_dir_path}/* ./ && rm -rf {tmp_dir_path}',
            # f'mv -f {tmp_dir_path}/* . && rm -rf {tmp_dir_path}',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

        stdout = subprocess.run(
            'git status',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info('Current status:', color='grey')
        print_info(stdout, color='grey')

        # Git add only modified and deleted
        print_info('Git add modified and commit and push it...')
        stdout = subprocess.run(
            'git add -u',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

        stdout = subprocess.run(
            'git status && git branch',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info('git status && git branch', color='grey')
        print_info(stdout, color='grey')

        stdout = subprocess.run(
            f'git commit -m "Export mr files" --allow-empty && git push --set-upstream private {private_branch_name}',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

    print_info('Deleting outdated private MR', color='pink')
    for mr in private_students_mr:
        print_info(f'Deleting outdated MR {mr.iid} {mr.title}...')
        # mr.state_event = 'close'
        # mr.save()
        mr.delete()


def create_public_mr(
        course_config: CourseConfig,
        object_attributes: dict[str, str] | None = None,
        *,
        dry_run: bool = False,
) -> None:
    """Copy changes from public repo"""
    raise NotImplementedError

    object_attributes = object_attributes or {}
    merge_commit_sha = object_attributes['merge_commit_sha']
    title = object_attributes['title']
    url = object_attributes['url']
    iid = object_attributes['iid']
    description = object_attributes['description']
    author_id = object_attributes['author_id']
    updated_at = object_attributes['updated_at']

    private_project = get_private_project(course_config.private_group, course_config.private_repo)
    full_private_project = GITLAB.projects.get(private_project.id)
    # public_project = get_public_project()
    # full_public_project = GITLAB.projects.get(public_project.id)
    author = GITLAB.users.get(author_id)

    # Get public project sha and generate branch name
    if merge_commit_sha:
        public_sha = merge_commit_sha
    else:
        public_sha = subprocess.run(
            f"git log public/{MASTER_BRANCH} --pretty=format:'%H' -n 1",
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
    print_info(f'public/{MASTER_BRANCH}: {title} (sha {public_sha})')
    new_branch_name = f'public/mr-{iid}'
    new_title = f'[PUBLIC] {title}'
    new_description = '  \n'.join([
        f'Merged public project [MR{iid}]({url})',
        f'author: {author.username} - {author.name}',
        f'updated_at: {updated_at}',
        f'sha: {public_sha}',
        '',
        description,
    ])

    # Get all mrs from private repo
    private_mrs = full_private_project.mergerequests.list(state='opened')
    private_mrs = {mr.title: mr for mr in private_mrs if '[PUBLIC]' in mr.title}

    if new_title in private_mrs:
        # Already exists
        mr = private_mrs[new_title]
        print_info(f'MR with sha {public_sha} already exists!')
        print_info(mr, color='grey')
        return
    else:
        # Not exist

        # Create branch
        print_info('Create branch from remote', color='grey')
        stdout = subprocess.run(
            f'git checkout --force -b {new_branch_name} private/{MASTER_BRANCH}',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

        print_info(f'Merge from public (by sha {public_sha[:7]})', color='grey')
        stdout = subprocess.run(
            # f'git merge -s ours --allow-unrelated-histories --no-commit public/{MASTER_BRANCH}',
            f'git merge --strategy-option=theirs --allow-unrelated-histories --no-commit {public_sha}',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

        # Check diff not empty
        stdout = subprocess.run(
            'git status',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

        if 'Changes to be committed:' not in stdout:
            print_info(
                f'Can not create MR. No changes to commit between {MASTER_BRANCH} and public/{MASTER_BRANCH}',
                color='orange'
            )
            return

        # Git add only modified and deleted
        print_info('Git add modified and commit and push it...')
        stdout = subprocess.run(
            f'git commit -m "{new_title}" && git push private {new_branch_name}',
            encoding='utf-8',
            shell=True, check=True,
            stdout=subprocess.PIPE,
        ).stdout
        print_info(stdout, color='grey')

        # Create mr
        mr = full_private_project.mergerequests.create({
            'source_branch': new_branch_name,
            'target_branch': MASTER_BRANCH,
            'title': new_title,
            'description': new_description,
            'labels': ['public'],
            'remove_source_branch': True,
            'squash': True,
            'allow_maintainer_to_push': True,
        })

        print_info('Ok. New MR created', color='green')
        print_info(mr.web_url)



from __future__ import annotations

from ..exceptions import PluginExecutionFailed
from .base import PluginABC, PluginOutput


class CheckRegexpsPlugin(PluginABC):
    """Plugin for checking forbidden regexps in a files."""

    name = "check_regexps"

    class Args(PluginABC.Args):
        origin: str
        patterns: list[str]
        regexps: list[str]
        # TODO: Add validation for patterns and regexps

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:  # type: ignore[override]
        # TODO: add verbose output with files list
        import re
        from pathlib import Path

        # TODO: move to Args validation
        if not Path(args.origin).exists():
            raise PluginExecutionFailed(
                f"Origin '{args.origin}' does not exist",
                output=f"Origin {args.origin} does not exist",
            )

        for pattern in args.patterns:
            for file in Path(args.origin).glob(pattern):
                if file.is_file():
                    with file.open() as f:
                        file_content = f.read()

                    for regexp in args.regexps:
                        if re.search(regexp, file_content, re.MULTILINE):
                            raise PluginExecutionFailed(
                                f"File '{file.name}' matches regexp '{regexp}'",
                                output=f"File '{file}' matches regexp '{regexp}'",
                            )
        return PluginOutput(
            output="No forbidden regexps found",
        )



from __future__ import annotations

from typing import Type

import pytest

from checker.configs import PipelineStageConfig
from checker.exceptions import BadConfig, PluginExecutionFailed
from checker.pipeline import PipelineRunner
from checker.plugins import PluginABC
from checker.plugins.base import PluginOutput


class _FailPlugin(PluginABC):
    name = "fail"

    def _run(self, args: PluginABC.Args, *, verbose: bool = False) -> PluginOutput:
        raise PluginExecutionFailed("Failed")


class _ScorePlugin(PluginABC):
    name = "score"

    class Args(PluginABC.Args):
        score: float = 0.5

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:
        if verbose:
            return PluginOutput(
                output=f"Score: {args.score:.2f}\nSome secret verbose line",
                percentage=args.score,
            )
        else:
            return PluginOutput(output=f"Score: {args.score:.2f}", percentage=args.score)


class _EchoPlugin(PluginABC):
    name = "echo"

    class Args(PluginABC.Args):
        message: str

    def _run(self, args: Args, *, verbose: bool = False) -> PluginOutput:
        return PluginOutput(output=args.message)


@pytest.fixture
def sample_plugins() -> dict[str, Type[PluginABC]]:
    return {
        "fail": _FailPlugin,
        "score": _ScorePlugin,
        "echo": _EchoPlugin,
    }


@pytest.fixture
def sample_correct_pipeline() -> list[PipelineStageConfig]:
    return [
        PipelineStageConfig(
            name="stage1 - echo",
            run="echo",
            args={"message": "${{ message }}"},
        ),
        PipelineStageConfig(
            name="stage2 - score",
            run="score",
            args={"score": 0.5},
            register_output="score_stage",
        ),
        PipelineStageConfig(
            name="stage3 - ignore fail",
            run="fail",
            fail="never",
        ),
        PipelineStageConfig(
            name="stage4 - skip fail if run_if=False",
            run="fail",
            run_if=False,
        ),
        PipelineStageConfig(
            name="stage5 - skip echo if run_if=False",
            run="echo",
            args={"message": "skipped message"},
            run_if=False,
        ),
        PipelineStageConfig(
            name="stage6 - skip fail if registered output",
            run="fail",
            run_if="${{ outputs.score_stage.percentage > 0.7 }}",
        ),
        PipelineStageConfig(
            name="stage7 - second echo",
            run="echo",
            args={"message": "second message"},
        ),
    ]


class TestSampleFixtures:
    def test_plugins(self, sample_plugins: dict[str, Type[PluginABC]]) -> None:
        plugin = sample_plugins["echo"]()
        plugin.validate({"message": "Hello"})
        result = plugin.run({"message": "Hello"}, verbose=True)
        assert result.percentage == 1.0
        assert result.output == "Hello"

        plugin = sample_plugins["score"]()
        plugin.validate({"score": 0.2})
        result = plugin.run({"score": 0.2})
        assert result.percentage == 0.2
        assert result.output == "Score: 0.20"

        plugin = sample_plugins["fail"]()
        plugin.validate({})
        with pytest.raises(PluginExecutionFailed):
            plugin.run({})


class TestPipelineRunnerValidation:
    def test_correct_pipeline_validation(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        pipeline_runner.validate({}, validate_placeholders=False)
        pipeline_runner.validate({"message": "Hello"}, validate_placeholders=True)
        with pytest.raises(BadConfig):
            pipeline_runner.validate({}, validate_placeholders=True)

    def test_unknown_plugin(self, sample_plugins: dict[str, Type[PluginABC]]) -> None:
        with pytest.raises(BadConfig) as exc_info:
            _ = PipelineRunner(
                pipeline=[
                    PipelineStageConfig(
                        name="stage1 - echo",
                        run="unknown",
                        args={"message": "Hello"},
                    ),
                ],
                plugins=sample_plugins,
                verbose=False,
            )
        assert "Unknown plugin" in str(exc_info.value)

    def test_validate_placeholders(self, sample_correct_pipeline: list[PipelineStageConfig]) -> None:
        with pytest.raises(BadConfig) as exc_info:
            _ = PipelineRunner(
                pipeline=sample_correct_pipeline,
                plugins={},
                verbose=False,
            )
        assert "Unknown plugin" in str(exc_info.value)

    def test_unknown_placeholder(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        with pytest.raises(BadConfig):
            pipeline_runner.validate({}, validate_placeholders=True)
        # TODO: fix it, now throwing Validation Error
        # assert "Unknown placeholder" in str(exc_info.value)

    def test_invalid_run_if(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        with pytest.raises(BadConfig):
            pipeline_runner.validate({"score": 0.5}, validate_placeholders=True)

    def test_invalid_register_output(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
    ) -> None:
        sample_correct_pipeline[1].register_output = "unknown"
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        with pytest.raises(BadConfig) as exc_info:
            pipeline_runner.validate({"message": "some valid message"}, validate_placeholders=True)
        assert "Invalid template" in str(exc_info.value)

    def test_run_correct_pipeline_verbose(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=True,
        )
        result = pipeline_runner.run({"message": "Hello"})
        assert not result.failed
        captured = capsys.readouterr().err
        assert "Hello" in captured
        assert "Score: 0.50" in captured
        assert "second message" in captured
        # in args print, so in the verbose output
        # assert "skipped message" not in captured.out and "skipped message" not in captured.err
        # verbose output
        assert "Some secret verbose line" in captured
        # stages names are printed
        for stage_name in ["stage1", "stage2", "stage3", "stage4", "stage5", "stage6"]:
            assert stage_name in captured

    def test_not_print_stage_if_not_verbose(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        result = pipeline_runner.run({"message": "Hello"})
        assert not result.failed
        captured = capsys.readouterr().err
        assert "stage4" not in captured
        assert "stage5" not in captured

    def test_run_correct_pipeline_not_verbose(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        result = pipeline_runner.run({"message": "Hello"})
        assert not result.failed
        captured = capsys.readouterr().err
        assert "Hello" in captured
        assert "Score: 0.50" in captured
        assert "second message" in captured
        # in args print, so not in the non-verbose output
        assert "skipped message" not in captured
        # verbose output
        assert "Some secret verbose line" not in captured
        # stages names are printed
        for stage_name in ["stage1", "stage2", "stage3"]:
            assert stage_name in captured

    def test_dry_run(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        result = pipeline_runner.run({"message": "Hello"}, dry_run=True)
        assert not result.failed
        captured = capsys.readouterr().err
        # no "error!" msg as all is skipped
        assert "error!" not in captured
        assert "[output here]" in captured
        assert "dry run!" in captured
        # stages names are printed
        for stage_name in ["stage1", "stage2", "stage3", "stage4", "stage5", "stage6"]:
            assert stage_name in captured

    def test_fail_fast(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        sample_correct_pipeline[2].fail = PipelineStageConfig.FailType.FAST
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        result = pipeline_runner.run({"message": "Hello"})
        assert result.failed
        captured = capsys.readouterr().err
        # no "error!" msg as all is skipped
        assert "error!" in captured
        # fist echo works, second not
        assert "Hello" in captured
        assert "second message" not in captured
        # stages names are printed
        for stage_name in ["stage1", "stage2", "stage3"]:
            assert stage_name in captured

    def test_fail_after_all(
        self,
        sample_correct_pipeline: list[PipelineStageConfig],
        sample_plugins: dict[str, Type[PluginABC]],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        sample_correct_pipeline[2].fail = PipelineStageConfig.FailType.AFTER_ALL
        pipeline_runner = PipelineRunner(
            pipeline=sample_correct_pipeline,
            plugins=sample_plugins,
            verbose=False,
        )
        result = pipeline_runner.run({"message": "Hello"})
        assert result.failed
        captured = capsys.readouterr().err
        # no "error!" msg as all is skipped
        assert "error!" in captured
        # fist echo works, second not
        assert "Hello" in captured
        assert "second message" in captured
        # stages names are printed
        for stage_name in ["stage1", "stage2", "stage3"]:
            assert stage_name in captured



from __future__ import annotations

import re
from datetime import datetime

import gitlab.v4.objects

from ..course import CourseConfig, CourseDriver
from ..course.schedule import CourseSchedule
from ..utils.glab import GitlabConnection
from ..utils.manytask import PushFailedError, push_report
from ..utils.print import print_header_info, print_info


BANNED_FILE_EXTENSIONS = {'csv', 'json', 'txt', 'db'}
ALLOWED_FILES = ['requirements.txt', 'runtime.txt']
REVIEWED_TAG = 'reviewed'
CHECKLIST_TAG = 'checklist'
BASIC_CHECKLIST_BANNED_TAGS = {CHECKLIST_TAG, REVIEWED_TAG}


def grade_students_mrs_to_master(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        private_course_driver: CourseDriver,
        gitlab_connection: GitlabConnection,
        *,
        dry_run: bool = False,
) -> None:
    students_projects = gitlab_connection.get_students_projects(course_config.students_group)
    # for project in students_projects:
    #     full_project: gitlab.v4.objects.Project = GITLAB.projects.get(project.id)
    #     username = full_project.name
    usernames = [project.name for project in students_projects]

    _grade_mrs(
        course_config,
        course_schedule,
        private_course_driver,
        gitlab_connection,
        usernames,
        dry_run=dry_run,
    )


def grade_student_mrs(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        course_driver: CourseDriver,
        gitlab_connection: GitlabConnection,
        username: str,
        *,
        dry_run: bool = False,
) -> None:
    _grade_mrs(
        course_config,
        course_schedule,
        course_driver,
        gitlab_connection,
        [username],
        dry_run=dry_run,
    )


def _grade_mrs(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        course_driver: CourseDriver,
        gitlab_connection: GitlabConnection,
        usernames: list[str],
        *,
        dry_run: bool = False,
) -> None:
    """
    Grade all users from list; to be used with individual and massive MRs check
    """

    # print users to check
    print_info('Users:', usernames, color='orange')

    # get open mrs to filter all users
    students_group = gitlab_connection.get_group(course_config.students_group)
    students_mrs: list[gitlab.v4.objects.GroupMergeRequest] = (students_group.mergerequests
                                                               .list(get_all=True))  # type: ignore
    students_mrs_project_names: set[str] = set()
    for mr in students_mrs:
        students_mrs_project_names.update(mr.web_url.split('/'))
    usernames = [i for i in usernames if i in students_mrs_project_names]
    print_info('Users with MRs:', usernames, color='orange')

    if len(usernames) == 0:
        print_info('Could not find MRs', color='orange')
        return

    # get tasks we need to check
    tag_to_folder = _get_tag_to_folder_dict(course_schedule, course_driver)
    print_info('Tags and folders to check:', tag_to_folder, color='orange')

    # get tutors
    tutors = gitlab_connection.get_all_tutors(course_config.course_group)
    print_info('Tutors:', [f'<{t.username} {t.name}>' for t in tutors], color='orange')
    id_to_tutor = {t.id: t for t in tutors}

    # get current user
    for username in usernames:
        try:
            user = gitlab_connection.get_user_by_username(username)
        except Exception:
            print_info(f'Can not find user with username={username}>', color='orange')
            continue

        user_id = user.id
        print_header_info(f'Current user: <{user.username} {user.name}>')

        # get current user's project
        project = gitlab_connection.get_project_from_group(course_config.students_group, user.username)
        full_project = gitlab_connection.gitlab.projects.get(project.id)
        print_info(f'project {project.path_with_namespace}: {project.web_url}')

        opened_master_mrs = full_project.mergerequests.list(
            get_all=True, state='opened', target_branch=course_config.default_branch,
        )
        merged_master_mrs = full_project.mergerequests.list(
            get_all=True, state='merged', target_branch=course_config.default_branch,
        )
        closed_master_mrs = full_project.mergerequests.list(
            get_all=True, state='closed', target_branch=course_config.default_branch,
        )

        if not opened_master_mrs and not merged_master_mrs and not closed_master_mrs:
            print_info('no open mrs; skip it')
            continue

        print_info(
            f'opened_master_mrs {len(opened_master_mrs)} \t '
            f'merged_master_mrs {len(merged_master_mrs)} \t '
            f'closed_master_mrs {len(closed_master_mrs)}',
            color='grey',
        )

        # Check basic checklist
        print_info('Lookup checklist')
        for mr in opened_master_mrs:  # type: ignore
            print_info(f'Checking MR#{mr.iid} <{mr.title}> ({mr.state})...', color='white')
            print_info(mr.web_url, color='white')
            if mr.title.lower().startswith('wip:') or mr.title.lower().startswith('draft:'):
                print_info('Draft MR - skip it.')
                continue
            _single_mr_check_basic_checklist(
                mr, tag_to_folder, dry_run=dry_run,
            )

        # Check score
        print_info('Lookup score to set')
        for mr in [*opened_master_mrs, *merged_master_mrs, *closed_master_mrs]:
            print_info(f'Checking MR#{mr.iid} <{mr.title}> ({mr.state})...', color='white')
            print_info(mr.web_url, color='white')
            if mr.title.lower().startswith('wip:') or mr.title.lower().startswith('draft:'):
                print_info('Draft MR - skip it.')
                continue
            _singe_mr_grade_score_new(
                course_config, course_schedule, mr, tag_to_folder, id_to_tutor, user_id, dry_run=dry_run,
            )


def _get_tag_to_folder_dict(course_schedule: CourseSchedule, course_driver: CourseDriver) -> dict[str, str]:
    tag_to_folder: dict[str, str] = {}
    for task in course_schedule.get_tasks(enabled=True):
        if task.review:
            source_dir = course_driver.get_task_dir(task)
            print_info(f'task "{task.name}" review=true with source_dir={source_dir}', color='grey')
            if source_dir is None:
                print_info('  source_dir is None, skip it', color='grey')
                continue
            tag_to_folder[task.name] = str(source_dir.relative_to(course_driver.root_dir))

    return tag_to_folder


def _singe_mr_grade_score_new(
        course_config: CourseConfig,
        course_schedule: CourseSchedule,
        mr: gitlab.v4.objects.GroupMergeRequest,
        tag_to_folder: dict[str, str],
        tutors_dict: dict[int, gitlab.v4.objects.GroupMember],
        user_id: int,
        *,
        dry_run: bool = False,
) -> None:
    """
    Get single MR, find or create score discussion and set a score from it
    Looking for comment by tutor under '#### MR score discussion:'
    """
    print_info('labels', mr.labels, color='grey')
    print_info('source_branch', mr.source_branch, color='grey')

    # Search for tags
    tag = None
    for search_tag in tag_to_folder:
        if search_tag in mr.labels:
            tag = search_tag
            break

    if not tag:
        print_info(f'Can not find any of {tag_to_folder.keys()} in MR tags ({mr.labels}). Skip it')
        return

    # get actual task
    task_name = tag_to_folder[tag].split('/')[-1]
    print_info('task_name', task_name, color='grey')
    max_task_score = course_schedule.tasks[task_name].max_score
    print_info('max_task_score', max_task_score, color='grey')

    # Try to find score discussion
    mr_score_discussion = None
    for discussion in mr.discussions.list(get_all=True):
        first_note_id = discussion.attributes['notes'][0]['id']
        first_note = discussion.notes.get(first_note_id)

        if '#### MR score discussion:' in first_note.body:
            mr_score_discussion = discussion
            break
    if not mr_score_discussion:
        mr_score_discussion = mr.discussions.create({
            'body': '\n  '.join([
                '#### MR score discussion:',
                '',
                'After review an examiner will put your score in a response to this discussion',
                f'than he/she will set `{REVIEWED_TAG}` label.',
                '',
                'If the score is not registered in the table, you need to restart the `grade-mr` job',
                '(last score accounted)'
            ])
        })
        try:
            mr_score_discussion.save()
        except Exception:
            print_info('ERROR with saving mr_score_discussion', color='orange')
        mr.save()

    if REVIEWED_TAG not in mr.labels:
        print_info(f'No `{REVIEWED_TAG}` tag. Skip it')
        return

    # get scores
    notes = [
        mr_score_discussion.notes.get(note['id'])
        for note in mr_score_discussion.attributes['notes']
    ]
    notes = notes[1:]

    if not notes:
        print_info('No replays on discussion note. Skip it.', color='grey')
        return

    if 'Score' in notes[-1].body and 'set' in notes[-1].body:
        print_info('Score already set. Skip it.', color='grey')
        return

    score_notes: list[tuple[int, gitlab.v4.objects.ProjectMergeRequestDiscussionNote]] = []
    for note in notes:
        if note.author['id'] not in tutors_dict:
            continue

        try:
            note_score = int(note.body)
        except Exception:
            continue

        if note.updated_at != note.created_at:
            print_info('Note was edited. Please, create a new one! Skip it.', color='grey')
            note.body = note.body + '\n  ' + '*(Note was edited. Please, create a new one! Skip it.)*'
            note.save()
            continue
        score_notes.append((note_score, note))

    if not score_notes:
        print_info('No score replays on discussion note. Skip it.', color='grey')
        return

    # set score from last score
    last_score, last_note = score_notes[-1]

    try:
        if not course_config.manytask_token:
            raise PushFailedError('Unable to find manytask token')

        username, score, _, _, _ = push_report(
            course_config.manytask_url,
            course_config.manytask_token,
            task_name,
            user_id,
            last_score,
            check_deadline=False,
            use_demand_multiplier=False,
        )
        print_info(
            f'Set score for @{username}: {score}',
            color='blue',
        )
        # print_info(f'Submit at {commit_time} (deadline is calculated relative to)', color='grey')
    except PushFailedError:
        raise

    mr_score_discussion.notes.create({'body': f'Score {last_score} set'})
    try:
        mr_score_discussion.save()
    except Exception:
        print_info('ERROR with saving mr_score_discussion', color='orange')
    print_info(f'Score {last_score} set', color='grey')


def _single_mr_check_basic_checklist(
        mr: gitlab.v4.objects.GroupMergeRequest,
        tag_to_folder: dict[str, str],
        *,
        dry_run: bool = False,
) -> None:
    print_info('pipelines', [i.status for i in mr.pipelines.list(get_all=True)], color='grey')
    print_info('labels', mr.labels, color='grey')
    print_info('source_branch', mr.source_branch, color='grey')

    # Search for tags
    tag = None
    for search_tag in tag_to_folder:
        if search_tag in mr.labels:
            tag = search_tag
            break

    if not tag:
        print_info(f'Can not find any of {tag_to_folder.keys()} in MR tags ({mr.labels}). Skip it')
        return

    for banned_tag in BASIC_CHECKLIST_BANNED_TAGS:
        if banned_tag in mr.labels:
            print_info(f'Have `{banned_tag}` tag. Skip it')
            return

    # Check status
    changes = mr.changes()
    is_conflict = mr.has_conflicts or mr.merge_status == 'cannot_be_merged'
    have_no_conflicts = not is_conflict

    # Check pipelines
    if not changes or not changes['head_pipeline']:
        head_pipeline_status = 'failed'
    else:
        head_pipeline_status = changes['head_pipeline']['status']
    pipeline_passed = head_pipeline_status == 'success'

    # changes files
    file_changed: set[str] = set()
    for change in changes['changes']:
        file_changed.update({change['old_path'], change['new_path']})
    print_info('file_changed', file_changed, color='grey')

    # Check single folder or not
    folder_prefix = tag_to_folder[tag]
    print_info('folder_prefix', folder_prefix, color='grey')
    is_single_folder = True
    wrong_folder = None
    for file in file_changed:
        if not file.startswith(folder_prefix):
            # print_info(f'file {file} not startswith {folder_prefix}', color='grey')
            is_single_folder = False
            wrong_folder = file
            break

    # Check extensions
    have_no_additional_files = True
    wrong_file = None
    for file in file_changed:
        if file.split('.')[-1] in BANNED_FILE_EXTENSIONS:
            print_info(f'  check {file}', color='grey')
            if any(file.endswith(allowed_file) for allowed_file in ALLOWED_FILES):
                continue
            have_no_additional_files = False
            wrong_file = file
            break

    # Get Mr checks discussions
    mr_checklist_discussion = None
    mr_checklist_note = None
    for discussion in mr.discussions.list(get_all=True):
        first_note_id = discussion.attributes['notes'][0]['id']
        note = discussion.notes.get(first_note_id)

        if '#### MR checklist (basic checks):' in note.body or '[MR check in progress...]' in note.body:
            mr_checklist_discussion = discussion
            mr_checklist_note = note
            break
    if not mr_checklist_discussion:
        mr_checklist_discussion = mr.discussions.create({'body': '[MR check in progress...]'})
        first_note_id = mr_checklist_discussion.attributes['notes'][0]['id']
        mr_checklist_note = mr_checklist_discussion.notes.get(first_note_id)

    assert mr_checklist_discussion
    assert mr_checklist_note

    # Generate note
    checks_ok = is_single_folder and have_no_additional_files and pipeline_passed and have_no_conflicts
    try:
        _first_try_correct_search = re.search(r'first try correct: (False|True)', mr_checklist_note.body)
        assert _first_try_correct_search
        _first_try_correct_str = _first_try_correct_search.group(0)
        _is_first_try_correct_search = re.search(r'(False|True)', _first_try_correct_str)
        assert _is_first_try_correct_search
        is_first_try_correct = _is_first_try_correct_search.group(0) == 'True'
    except (ValueError, AttributeError, AssertionError):
        is_first_try_correct = checks_ok
    try:
        _updates_num_search = re.search(r'checks num: (\d+)', mr_checklist_note.body)
        assert _updates_num_search
        _updates_num_str = _updates_num_search.group(0)
        current_updates_num_search = re.search(r'(\d+)', _updates_num_str)
        assert current_updates_num_search
        current_updates_num = int(current_updates_num_search.group(0))
    except (ValueError, AttributeError, AssertionError):
        current_updates_num = 0
    now_str = str(datetime.now())
    checklist_note_msg = [
        '#### MR checklist (basic checks):',
        '',
        'Hi! I\'m course bot; I\'m here to check your merge request.',
        'Below you will find a checklist to verify your MR.',
        '',
        f'_first try correct: {is_first_try_correct}_',
        f'_checks num: {current_updates_num + 1}_',
        f'_last check time: {now_str}_',
        '',
        f'- [x] `{tag}` tag exists',
        f'- [{"x" if is_single_folder else " "}] all changes in single folder '
        f'{f"(found {wrong_folder})" if not is_single_folder else ""}',
        f'- [{"x" if have_no_additional_files else " "}] no additional files '
        f'{f"(found {wrong_file})" if not have_no_additional_files else ""}',
        f'- [{"x" if pipeline_passed else " "}] pipeline passed (current status: {head_pipeline_status})',
        f'- [{"x" if have_no_conflicts else " "}] have no merge conflicts',
        '',
    ]

    # TODO: fix it
    if tag == 'cinemabot':
        have_bot_tag = '@' in mr.description
        checks_ok = checks_ok and have_bot_tag
        checklist_note_msg.insert(-1, f'- [{"x" if have_bot_tag else " "}] placed @bot_tag in description')

    # Update MR check discussion
    if checks_ok:
        checklist_note_msg.append('💪 **Ok.** Basic checks have been passed!')
        checklist_note_msg.append('Please, wait for a examiner to check it manually.')
        mr_checklist_discussion.resolved = True
        mr.labels = list({*mr.labels, 'checklist'})
    else:
        checklist_note_msg.append('🔥 **Oops!** There are some errors;')
        checklist_note_msg.append('Please, correct it.')
        mr_checklist_discussion.resolved = False
        mr.labels = list({*mr.labels, 'fix it'})
    # print_info('  \n  '.join(checklist_note_msg))
    mr_checklist_note.body = '  \n'.join(checklist_note_msg)
    mr_checklist_note.save()
    mr_checklist_discussion.save()
    mr.save()

    print_info('Checklist updated')



from .files import *  # noqa: F403
from .git import *  # noqa: F403
from .glab import *  # noqa: F403
from .manytask import *  # noqa: F403
from .print import *  # noqa: F403
from .template import *  # noqa: F403



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from .configs.checker import CheckerConfig, CheckerParametersConfig
from .course import Course, FileSystemTask
from .exceptions import TestingError
from .pipeline import PipelineResult, PipelineRunner, PipelineStageResult
from .plugins import load_plugins
from .utils import print_header_info, print_info, print_separator


@dataclass
class GlobalPipelineVariables:
    """Base variables passed in pipeline stages."""

    ref_dir: str
    repo_dir: str
    temp_dir: str
    task_names: list[str]
    task_sub_paths: list[str]


@dataclass
class TaskPipelineVariables:
    """Variables passed in pipeline stages for each task."""

    task_name: str
    task_sub_path: str


class Tester:
    """
    Class to encapsulate all testing logic.
    1. Accept directory with files ready for testing
    2. Execute global pipeline once
    3. For each task:
        3.1. Execute task pipeline
        3.2. Execute report pipeline (optional)
    """

    __test__ = False  # do not collect this class for pytest

    def __init__(
        self,
        course: Course,
        checker_config: CheckerConfig,
        *,
        verbose: bool = False,
        dry_run: bool = False,
    ):
        """
        Init tester in specific public and private dirs.

        :param course: Course object for iteration with physical course
        :param checker_config: Full checker config with testing,structure and params folders
        :param verbose: Whatever to print private outputs and debug info
        :param dry_run: Do not execute anything, just print what would be executed
        :raises exception.ValidationError: if config is invalid or repo structure is wrong
        """
        self.course = course

        self.testing_config = checker_config.testing
        self.structure_config = checker_config.structure
        self.default_params = checker_config.default_parameters

        self.plugins = load_plugins(self.testing_config.search_plugins, verbose=verbose)

        self.global_pipeline = PipelineRunner(self.testing_config.global_pipeline, self.plugins, verbose=verbose)
        self.task_pipeline = PipelineRunner(self.testing_config.tasks_pipeline, self.plugins, verbose=verbose)
        self.report_pipeline = PipelineRunner(self.testing_config.report_pipeline, self.plugins, verbose=verbose)

        self.repository_dir = self.course.repository_root
        self.reference_dir = self.course.reference_root

        self.verbose = verbose
        self.dry_run = dry_run

    def _get_global_pipeline_parameters(
        self,
        origin: Path,
        tasks: list[FileSystemTask],
    ) -> GlobalPipelineVariables:
        return GlobalPipelineVariables(
            ref_dir=self.reference_dir.absolute().as_posix(),
            repo_dir=self.repository_dir.absolute().as_posix(),
            temp_dir=origin.absolute().as_posix(),
            task_names=[task.name for task in tasks],
            task_sub_paths=[task.relative_path for task in tasks],
        )

    def _get_task_pipeline_parameters(
        self,
        task: FileSystemTask,
    ) -> TaskPipelineVariables:
        return TaskPipelineVariables(
            task_name=task.name,
            task_sub_path=task.relative_path,
        )

    def _get_context(
        self,
        global_variables: GlobalPipelineVariables,
        task_variables: TaskPipelineVariables | None,
        outputs: dict[str, PipelineStageResult],
        default_parameters: CheckerParametersConfig,
        task_parameters: CheckerParametersConfig | None,
    ) -> dict[str, Any]:
        return {
            "global": global_variables,
            "task": task_variables,
            "outputs": outputs,
            "parameters": default_parameters.__dict__ | (task_parameters.__dict__ if task_parameters else {}),
            "env": os.environ.__dict__,
        }

    def validate(self) -> None:
        # get all tasks
        tasks = self.course.get_tasks(enabled=True)

        # create outputs to pass to pipeline
        outputs: dict[str, PipelineStageResult] = {}

        # validate global pipeline (only default params and variables available)
        print_info("- global pipeline...")
        global_variables = self._get_global_pipeline_parameters(Path(), tasks)
        context = self._get_context(global_variables, None, outputs, self.default_params, None)
        self.global_pipeline.validate(context, validate_placeholders=True)
        print_info("  ok")

        for task in tasks:
            # validate task with global + task-specific params
            print_info(f"- task {task.name} pipeline...")

            # create task context
            task_variables = self._get_task_pipeline_parameters(task)
            context = self._get_context(
                global_variables,
                task_variables,
                outputs,
                self.default_params,
                task.config.parameters if task.config else None,
            )

            # check task parameter are
            # TODO: read pipeline from task config if any
            self.task_pipeline.validate(context, validate_placeholders=True)
            self.report_pipeline.validate(context, validate_placeholders=True)

            print_info("  ok")

    def run(
        self,
        origin: Path,
        tasks: list[FileSystemTask] | None = None,
        report: bool = True,
    ) -> None:
        # get all tasks
        tasks = tasks or self.course.get_tasks(enabled=True)

        # create outputs to pass to pipeline
        outputs: dict[str, PipelineStageResult] = {}

        # run global pipeline
        global_variables = self._get_global_pipeline_parameters(origin, tasks)
        if len(self.global_pipeline) > 0 or self.verbose:
            print_header_info("Run global pipeline:", color="pink")
            context = self._get_context(
                global_variables,
                None,
                outputs,
                self.default_params,
                None,
            )
            global_pipeline_result: PipelineResult = self.global_pipeline.run(context, dry_run=self.dry_run)
            print_separator("-")
            print_info(str(global_pipeline_result), color="pink")

            if not global_pipeline_result:
                raise TestingError("Global pipeline failed")

        failed_tasks = []
        for task in tasks:
            # run task pipeline
            print_header_info(f"Run <{task.name}> task pipeline:", color="pink")

            # create task context
            task_variables = self._get_task_pipeline_parameters(task)
            context = self._get_context(
                global_variables,
                task_variables,
                outputs,
                self.default_params,
                task.config.parameters if task.config else None,
            )

            # TODO: read pipeline from task config if any
            task_pipeline_result: PipelineResult = self.task_pipeline.run(context, dry_run=self.dry_run)
            print_separator("-")

            print_info(str(task_pipeline_result), color="pink")
            print_separator("-")

            # Report score if task pipeline succeeded
            if task_pipeline_result:
                print_info(f"Reporting <{task.name}> task tests:", color="pink")
                if report:
                    task_report_result: PipelineResult = self.report_pipeline.run(context, dry_run=self.dry_run)
                    if task_report_result:
                        print_info("->Reporting succeeded")
                    else:
                        print_info("->Reporting failed")
                else:
                    _: PipelineResult = self.report_pipeline.run(context, dry_run=True)
                    print_info("->Reporting disabled (dry-run)")
                print_separator("-")
            else:
                failed_tasks.append(task.name)

        if failed_tasks:
            raise TestingError(f"Task pipelines failed: {failed_tasks}")



from __future__ import annotations

from enum import Enum
from typing import Optional, Union

from pydantic import AnyUrl, Field, RootModel, ValidationError, field_validator

from .utils import CustomBaseModel, YamlLoaderMixin


# Note: old Union style in definition for backward compatibility
TParamType = Union[bool, int, float, str, list[Union[int, float, str, None]], None]
TTemplate = Union[str, list[Union[TParamType, str]], dict[str, Union[TParamType, str]]]


class CheckerStructureConfig(CustomBaseModel):
    # Note: use Optional/Union[...] instead of ... | None as pydantic does not support | in older python versions
    ignore_patterns: Optional[list[str]] = None
    private_patterns: Optional[list[str]] = None
    public_patterns: Optional[list[str]] = None
    # TODO: add check "**" is not allowed


class CheckerParametersConfig(RootModel[dict[str, TParamType]]):
    root: dict[str, TParamType]

    def __getitem__(self, item: str) -> TParamType:
        return self.root[item]

    def __contains__(self, item: str) -> bool:
        return item in self.root

    @property
    def __dict__(self) -> dict[str, TParamType]:
        return self.root

    @__dict__.setter
    def __dict__(self, value: dict[str, TParamType]) -> None:
        self.root = value


class CheckerExportConfig(CustomBaseModel):
    class TemplateType(Enum):
        """Template type for export for each task.

        :ivar SEARCH: search for files/folder with name `some_file.template` and override `some_file` with it.
            If `some_file.template` is empty file/folder it will delete `some_file`.
            For ALL `some_file.template` files original `some_file` HAVE TO exist.
            At least one `some_file.template` HAVE TO exist for each task.
        :ivar CREATE: for all files in the repo search for template comments and delete all code between them.
            For example:
            ```python
            a = 1
            # SOLUTION BEGIN
            a = 2
            # SOLUTION END
            b = 3
            ```
            will be converted to:
            ```python
            a = 1
            # TODO: Your solution
            b = 3
            ```
            Delete file if it is empty after template comments deletion.
            Each task HAVE to contain at least one template comment pair.
        :ivar SEARCH_OR_CREATE: try to search for files/folder with name `some_file.template`
            if not found try to create it using template comments.
        """

        SEARCH = "search"
        CREATE = "create"
        SEARCH_OR_CREATE = "search_or_create"

    destination: AnyUrl
    default_branch: str = "main"
    commit_message: str = "chore(auto): export new tasks"
    templates: TemplateType = TemplateType.SEARCH


class PipelineStageConfig(CustomBaseModel):
    class FailType(Enum):
        FAST = "fast"
        AFTER_ALL = "after_all"
        NEVER = "never"

    name: str
    run: str

    # Note: use Optional/Union[...] instead of ... | None as pydantic does not support | in older python versions
    args: dict[str, Union[TParamType, TTemplate]] = Field(default_factory=dict)

    run_if: Union[bool, TTemplate, None] = None
    fail: FailType = FailType.FAST

    # save pipline stage result to context under this key
    register_output: Optional[str] = None


class CheckerTestingConfig(CustomBaseModel):
    class ChangesDetectionType(Enum):
        BRANCH_NAME = "branch_name"
        COMMIT_MESSAGE = "commit_message"
        LAST_COMMIT_CHANGES = "last_commit_changes"
        FILES_CHANGED = "files_changed"

    changes_detection: ChangesDetectionType = ChangesDetectionType.LAST_COMMIT_CHANGES

    search_plugins: list[str] = Field(default_factory=list)

    global_pipeline: list[PipelineStageConfig] = Field(default_factory=list)
    tasks_pipeline: list[PipelineStageConfig] = Field(default_factory=list)
    report_pipeline: list[PipelineStageConfig] = Field(default_factory=list)


class CheckerConfig(CustomBaseModel, YamlLoaderMixin["CheckerConfig"]):
    """
    Checker configuration.

    :ivar version: config version
    :ivar default_parameters: default parameters for task pipeline
    :ivar structure: describe the structure of the repo - private/public and allowed for change files
    :ivar export: describe export (publishing to public repo)
    :ivar manytask: describe connection to manytask
    :ivar testing: describe testing/checking - pipeline, isolation etc
    """

    version: int  # if config exists, version is always present

    default_parameters: CheckerParametersConfig = Field(default_factory=dict)

    structure: CheckerStructureConfig
    export: CheckerExportConfig
    testing: CheckerTestingConfig

    @field_validator("version")
    @classmethod
    def check_version(cls, v: int) -> None:
        if v != 1:
            raise ValidationError(f"Only version 1 is supported for {cls.__name__}")


class CheckerSubConfig(CustomBaseModel, YamlLoaderMixin["CheckerSubConfig"]):
    """
    Configuration file overwriting CheckerConfig for some folder and subfolders

    :ivar version: config version
    :ivar structure: describe the structure of the repo - private/public and allowed for change files
    :ivar parameters: parameters for task pipeline
    :ivar task_pipeline: pipeline run for each task
    :ivar report_pipeline: pipeline run for each task if task_pipeline succeeded
    """

    version: int  # if config exists, version is always present

    # Note: use Optional[...] instead of ... | None as pydantic does not support | in older python versions
    structure: Optional[CheckerStructureConfig] = None
    parameters: Optional[CheckerParametersConfig] = None
    task_pipeline: Optional[list[PipelineStageConfig]] = None
    report_pipeline: Optional[list[PipelineStageConfig]] = None

    @classmethod
    def default(cls) -> "CheckerSubConfig":
        return CheckerSubConfig(version=1)

    @field_validator("version")
    @classmethod
    def check_version(cls, v: int) -> None:
        if v != 1:
            raise ValidationError(f"Only version 1 is supported for {cls.__name__}")



from __future__ import annotations

import re
import shutil
import tempfile
from pathlib import Path

from checker.configs import CheckerExportConfig, CheckerStructureConfig
from checker.course import Course
from checker.exceptions import BadStructure
from checker.utils import print_info


class Exporter:
    """
    The Exporter class is responsible for moving course files.
    1. It validates and manage templates
    2. Select files to be exported public/testing/contribution
    """

    TEMPLATE_SUFFIX = ".template"
    TEMPLATE_START_COMMENT = "SOLUTION BEGIN"
    TEMPLATE_END_COMMENT = "SOLUTION END"
    TEMPLATE_REPLACE_COMMENT = "TODO: Your solution"
    TEMPLATE_COMMENT_REGEX = re.compile(f"{TEMPLATE_START_COMMENT}(.*?){TEMPLATE_END_COMMENT}", re.DOTALL)

    def __init__(
        self,
        course: Course,
        structure_config: CheckerStructureConfig,
        export_config: CheckerExportConfig,
        *,
        cleanup: bool = True,
        verbose: bool = False,
        dry_run: bool = False,
    ) -> None:
        self.course = course

        self.structure_config = structure_config
        self.export_config = export_config

        self.repository_root = course.repository_root
        self.reference_root = course.reference_root

        self._temporary_dir_manager = tempfile.TemporaryDirectory()
        self.temporary_dir = Path(self._temporary_dir_manager.name)

        self.sub_config_files = {}
        for group in self.course.get_groups(enabled=True):
            relative_path = Path(group.relative_path)
            if group.config.structure:
                self.sub_config_files[relative_path] = group.config.structure
        for task in self.course.get_tasks(enabled=True):
            relative_path = Path(task.relative_path)
            if task.config.structure:
                self.sub_config_files[relative_path] = task.config.structure

        self.cleanup = cleanup
        self.verbose = verbose
        self.dry_run = dry_run

    def validate(self) -> None:
        # validate course
        self.course.validate()

        # TODO: validate structure correct glob patterns

        # validate templates
        # `original.template` files/folders need to have `original` file/folder
        # template comments have to be paired
        # template comments can not be one inside another
        # if `templates` is set to `SEARCH` - only `.template` allowed
        # if `templates` is set to `CREATE` - only template comments allowed
        # if `templates` is set to `SEARCH_OR_CREATE` - both allowed, but one inside one task
        for task in self.course.get_tasks(enabled=True):
            # TODO: check template not public and not private file

            task_folder = self.reference_root / task.relative_path
            task_has_template_files, task_has_valid_template_files = False, False
            task_has_template_comments, task_has_valid_template_comments = False, False

            # search for all `.template` files or folders
            for template_file_or_folder in task_folder.glob(f"**/*{self.TEMPLATE_SUFFIX}"):
                task_has_template_files = True
                # check that all files have original files
                if not (template_file_or_folder.parent / template_file_or_folder.stem).exists():
                    raise BadStructure(
                        f"Template file/folder {template_file_or_folder} does not have "
                        f"original file/folder {self.reference_root / template_file_or_folder.stem}"
                    )
                task_has_valid_template_files = True

            # check all (not binary) files for template comments
            for potential_comments_file in task_folder.glob("**/*"):
                if potential_comments_file.is_dir():
                    continue
                with potential_comments_file.open("r") as f:
                    # skip binary files
                    try:
                        file_content = f.read()
                    except UnicodeDecodeError:
                        continue

                    # validate using regex and count matches of start and end comments
                    if self.TEMPLATE_START_COMMENT in file_content or self.TEMPLATE_END_COMMENT in file_content:
                        task_has_template_comments = True

                        # check have equal num of comments
                        if file_content.count(self.TEMPLATE_START_COMMENT) != file_content.count(
                            self.TEMPLATE_END_COMMENT
                        ):
                            task_has_valid_template_comments = False
                            raise BadStructure(
                                f"Task {task.name} has invalid template comments in file {potential_comments_file}. "
                                f"The number of <{self.TEMPLATE_START_COMMENT}> and "
                                f"<{self.TEMPLATE_END_COMMENT}> do not match"
                            )
                        # check between comments no other comment pair
                        for match in self.TEMPLATE_COMMENT_REGEX.finditer(file_content):
                            if self.TEMPLATE_START_COMMENT in match.group(
                                1
                            ) or self.TEMPLATE_END_COMMENT in match.group(1):
                                task_has_valid_template_comments = False
                                raise BadStructure(
                                    f"Task {task.name} has invalid template comments in file {potential_comments_file}."
                                    f" There is <{self.TEMPLATE_START_COMMENT}> or <{self.TEMPLATE_END_COMMENT}> "
                                    f"between valid pair of comments"
                                )

                        task_has_valid_template_comments = True

            if self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH:
                if task_has_template_comments:
                    raise BadStructure(
                        f"Templating set to {self.export_config.templates} but task {task.name} has "
                        f"template comments <{self.TEMPLATE_START_COMMENT}> and <{self.TEMPLATE_END_COMMENT}>"
                    )
                if not task_has_valid_template_files:
                    raise BadStructure(
                        f"Task {task.name} does not have `.template` file/folder. Have to include at least one"
                    )
            elif self.export_config.templates == CheckerExportConfig.TemplateType.CREATE:
                if task_has_template_files:
                    raise BadStructure(
                        f"Templating set to {self.export_config.templates} but task {task.name} has "
                        f"`.template` file/folder"
                    )
                if not task_has_valid_template_comments:
                    raise BadStructure(
                        f"Task {task.name} does not have template comments. Have to include at least one pair of "
                        f"<{self.TEMPLATE_START_COMMENT}> and <{self.TEMPLATE_END_COMMENT}>"
                    )
            elif self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH_OR_CREATE:
                if task_has_template_files and task_has_template_comments:
                    raise BadStructure(
                        f"Task {task.name} can not use both `.template` file/folder and template comments"
                    )
                if not task_has_valid_template_files and not task_has_valid_template_comments:
                    raise BadStructure(
                        f"Task {task.name} does not have `.template` file/folder or at least one pair of "
                        f"<{self.TEMPLATE_START_COMMENT}> and <{self.TEMPLATE_END_COMMENT}>"
                    )
            else:  # pragma: no cover
                assert False, "Not Reachable"

    def _search_for_exclude_due_to_templates(
        self,
        root: Path,
        ignore_templates: bool,
    ) -> list[str]:
        """Search for files/folder should be ignored due to templating in the current directory only"""
        exclude_paths = []

        if (
            self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH
            or self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH_OR_CREATE
        ):
            for template_file_or_folder in root.glob(f"*{self.TEMPLATE_SUFFIX}"):
                if ignore_templates:
                    exclude_paths.append(template_file_or_folder.name)
                else:
                    exclude_paths.append(template_file_or_folder.stem)

        if (
            self.export_config.templates == CheckerExportConfig.TemplateType.CREATE
            or self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH_OR_CREATE
        ):
            # if got empty file after template comments deletion - exclude it
            for potential_comments_file in root.glob("*"):
                if potential_comments_file.is_dir():
                    continue
                try:
                    open(potential_comments_file, "r").read()
                except UnicodeDecodeError:
                    continue
                with potential_comments_file.open("r") as f:
                    file_content = f.read().strip()
                    if file_content.startswith(self.TEMPLATE_START_COMMENT) and file_content.endswith(
                        self.TEMPLATE_END_COMMENT
                    ):
                        exclude_paths.append(potential_comments_file.name)

        return exclude_paths

    def export_public(
        self,
        target: Path,
        push: bool = False,
        commit_message: str = "chore(auto): Update public files [skip-ci]",
    ) -> None:
        target.mkdir(parents=True, exist_ok=True)

        disabled_groups_and_tasks_to_skip = [
            *[group.relative_path for group in self.course.get_groups(enabled=False)],
            *[group.relative_path for group in self.course.get_groups(started=False)],
            *[task.relative_path for task in self.course.get_tasks(enabled=False)],
            *[task.relative_path for task in self.course.get_tasks(started=False)],
        ]

        print_info(f"Copy from {self.reference_root} to {target}", color="grey")
        self._copy_files_with_config(
            self.reference_root,
            target,
            self.structure_config,
            copy_public=True,
            copy_private=False,
            copy_other=True,
            fill_templates=True,
            extra_ignore_paths=disabled_groups_and_tasks_to_skip,
        )

    def export_for_testing(
        self,
        target: Path,
    ) -> None:
        target.mkdir(parents=True, exist_ok=True)

        print_info(f"Copy from {self.repository_root} to {target}", color="grey")
        self._copy_files_with_config(
            self.repository_root,
            target,
            self.structure_config,
            copy_public=False,
            copy_private=False,
            copy_other=True,
            fill_templates=False,
        )

        print_info(f"Copy from {self.reference_root} to {target}", color="grey")
        self._copy_files_with_config(
            self.reference_root,
            target,
            self.structure_config,
            copy_public=True,
            copy_private=True,
            copy_other=False,
            fill_templates=False,
        )

    def export_for_contribution(
        self,
        target: Path,
    ) -> None:
        target.mkdir(parents=True, exist_ok=True)

        print_info(f"Copy from {self.repository_root} to {target}", color="grey")
        self._copy_files_with_config(
            self.repository_root,
            target,
            self.structure_config,
            copy_public=True,
            copy_private=False,
            copy_other=True,
            fill_templates=False,
        )

        print_info(f"Copy from {self.reference_root} to {target}", color="grey")
        self._copy_files_with_config(
            self.reference_root,
            target,
            self.structure_config,
            copy_public=False,
            copy_private=True,
            copy_other=True,
            fill_templates=False,
        )

    def _copy_files_with_config(
        self,
        root: Path,
        destination: Path,
        config: CheckerStructureConfig,
        copy_public: bool,
        copy_private: bool,
        copy_other: bool,
        fill_templates: bool,
        extra_ignore_paths: list[str] | None = None,
        global_root: Path | None = None,
        global_destination: Path | None = None,
    ) -> None:
        """
        Copy files from `root` to `destination` according to `config`.
        When face `sub_config_files`, apply it to the folder and all subfolders.

        :param root: Copy files from this directory
        :param destination: Copy files to this directory
        :param config: Config to apply to this folder (and recursively)
        :param copy_public: Copy public files
        :param copy_private: Copy private files
        :param copy_other: Copy other - not public and not private files
        :param fill_templates: Fill templates (`.template` or template comments), if false will delete them
        :param extra_ignore_paths: Extra paths to ignore to skip not-enables groups/tasks, relative to `global_root`
        :param global_root: Starting root directory
        :param global_destination: Starting destination directory
        """
        # TODO: implement template searcher

        global_root = global_root or root
        global_destination = global_destination or destination

        if self.verbose:
            print_info(
                f"Copy files from <{root.relative_to(global_root)}> to <{destination.relative_to(global_destination)}>",
                color="white",
            )
            print_info(f"  {config=}", color="white")

        if extra_ignore_paths is not None:
            if str(root.relative_to(global_root)) in extra_ignore_paths:
                if self.verbose:
                    print_info(
                        f"    - Skip <{root.relative_to(global_root)}> because of extra ignore paths", color="grey"
                    )
                return

        # select paths to ignore - original to replace or templates to ignore
        exclude_paths = self._search_for_exclude_due_to_templates(root, not fill_templates)

        # Iterate over all files in the root directory
        for path in root.iterdir():
            path_destination = destination / path.relative_to(root)
            # check if byte file
            is_text_file = False
            try:
                if path.is_file():
                    open(path, "r").read()
                    is_text_file = True
            except UnicodeDecodeError:
                pass
            # check if file template
            is_path_template_file = (
                self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH
                or self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH_OR_CREATE
            ) and path.name.endswith(self.TEMPLATE_SUFFIX)
            is_path_template_comment = (
                is_text_file
                and (
                    self.export_config.templates == CheckerExportConfig.TemplateType.CREATE
                    or self.export_config.templates == CheckerExportConfig.TemplateType.SEARCH_OR_CREATE
                )
                and not path.is_dir()
                and self.TEMPLATE_START_COMMENT in path.read_text()
                and self.TEMPLATE_END_COMMENT in path.read_text()
            )

            # if will replace with template - ignore file
            if path.name in exclude_paths:
                if self.verbose:
                    print_info(f"    - Skip <{path.relative_to(global_root)}> because of templating", color="grey")
                continue

            # ignore if match ignore patterns
            if config.ignore_patterns and any(path.match(ignore_pattern) for ignore_pattern in config.ignore_patterns):
                if self.verbose:
                    print_info(f"    - Skip <{path.relative_to(global_root)}> because of ignore patterns", color="grey")
                continue

            # If matches public patterns AND copy_public is False - skip
            is_public = False
            if config.public_patterns and any(path.match(public_pattern) for public_pattern in config.public_patterns):
                is_public = True
                if not copy_public:
                    if self.verbose:
                        print_info(
                            f"    - Skip <{path.relative_to(global_root)}> because of public patterns skip",
                            color="grey",
                        )
                    continue

            # If matches private patterns AND copy_private is False - skip
            # If it is public file - never consider it as private
            is_private = False
            if (
                not is_public
                and config.private_patterns
                and any(path.match(private_pattern) for private_pattern in config.private_patterns)
            ):
                is_private = True
                if not copy_private:
                    if self.verbose:
                        print_info(
                            f"    - Skip <{path.relative_to(global_root)}> because of skip private patterns skip",
                            color="grey",
                        )
                    continue

            # if not match public and not match private and copy_other is False - skip
            # Note: never skip "other" directories, look inside them first
            if not is_public and not is_private and not path.is_dir():
                if not copy_other:
                    if self.verbose:
                        print_info(
                            f"    - Skip <{path.relative_to(global_root)}> because of copy other files not enabled",
                            color="grey",
                        )
                    continue

            # if file is empty file/folder - just do not copy (delete original file due to exclude_paths)
            if fill_templates and is_path_template_file:
                if path.is_dir() and not any((path_destination / file).exists() for file in path.iterdir()):
                    if self.verbose:
                        print_info(
                            f"    - Skip <{path.relative_to(global_root)}> because it is empty folder and "
                            f"templating is set to {self.export_config.templates}",
                            color="grey",
                        )
                    continue
                if path.is_file() and path.stat().st_size == 0:
                    if self.verbose:
                        print_info(
                            f"    - Skip <{path.relative_to(global_root)}> because it is empty file and "
                            f"templating is set to {self.export_config.templates}",
                            color="grey",
                        )
                    continue

            # If the file is a directory, recursively call this function
            if path.is_dir():
                # if folder public or private - just copy it
                if is_public or is_private:
                    if self.verbose:
                        print_info(
                            f"    - Fully Copy <{path.relative_to(global_root)}> to "
                            f"<{path_destination.relative_to(global_destination)}>",
                            color="grey",
                        )
                    self._copy_files_with_config(
                        path,
                        path_destination,
                        config,
                        copy_public=True,
                        copy_private=True,
                        copy_other=True,
                        fill_templates=fill_templates,
                        extra_ignore_paths=extra_ignore_paths,
                        global_root=global_root,
                        global_destination=global_destination,
                    )
                    continue

                # If directory `origin.template` - copy from this folder to `origin`
                if fill_templates and is_path_template_file:
                    path_destination = path_destination.parent / path_destination.stem

                # If have sub-config - update config with sub-config
                if path.relative_to(global_root) in self.sub_config_files:
                    declared_sub_config = self.sub_config_files[path.relative_to(global_root)]
                    sub_config = CheckerStructureConfig(
                        ignore_patterns=declared_sub_config.ignore_patterns
                        if declared_sub_config.ignore_patterns is not None
                        else config.ignore_patterns,
                        private_patterns=declared_sub_config.private_patterns
                        if declared_sub_config.private_patterns is not None
                        else config.private_patterns,
                        public_patterns=declared_sub_config.public_patterns
                        if declared_sub_config.public_patterns is not None
                        else config.public_patterns,
                    )
                else:
                    sub_config = config

                # Recursively call this function
                if self.verbose:
                    print_info(
                        f"    -- Recursively copy from <{path.relative_to(global_root)}> to "
                        f"<{path_destination.relative_to(global_destination)}>",
                        color="grey",
                    )
                self._copy_files_with_config(
                    path,
                    path_destination,
                    sub_config,
                    copy_public,
                    copy_private,
                    copy_other,
                    fill_templates,
                    extra_ignore_paths,
                    global_root=global_root,
                    global_destination=global_destination,
                )
            # If the file is a normal file, copy it
            else:
                if self.verbose:
                    print_info(
                        f"    - Copy <{path.relative_to(global_root)}> to "
                        f"<{path_destination.relative_to(global_destination)}>",
                        color="grey",
                    )
                path_destination.parent.mkdir(parents=True, exist_ok=True)

                # if `origin.template` - copy from this file as `origin`
                if fill_templates and is_path_template_file:
                    path_destination = path_destination.parent / path_destination.stem

                # if template comments in file - replace them, not greedy
                if fill_templates and is_path_template_comment:
                    with path.open("r") as f:
                        file_content = f.read()
                    file_content = self.TEMPLATE_COMMENT_REGEX.sub(self.TEMPLATE_REPLACE_COMMENT, file_content)
                    path_destination.touch(exist_ok=True)
                    path_destination.write_text(file_content)
                else:
                    shutil.copyfile(
                        path,
                        path_destination,
                    )

    def __del__(self) -> None:
        if self.__dict__.get("cleanup") and self._temporary_dir_manager:
            self._temporary_dir_manager.cleanup()



from __future__ import annotations

import typing
from pathlib import Path

import gitlab
import gitlab.v4.objects

from .print import print_info


class GitlabConnection:
    def __init__(
            self,
            gitlab_host_url: str,
            api_token: str | None = None,
            private_token: str | None = None,
            job_token: str | None = None,
    ):
        if api_token:
            self.gitlab = gitlab.Gitlab(gitlab_host_url, private_token=api_token)
        elif private_token:
            self.gitlab = gitlab.Gitlab(gitlab_host_url, private_token=private_token)
        elif job_token:
            self.gitlab = gitlab.Gitlab(gitlab_host_url, job_token=job_token)
        else:
            print_info(
                'None of `api_token`/`private_token` or `job_token` provided; use without credentials',
                color='orange',
            )
            self.gitlab = gitlab.Gitlab(gitlab_host_url)

    def get_project_from_group(
            self,
            group_name: str,
            project_name: str,
    ) -> gitlab.v4.objects.GroupProject:
        print_info('Get private Project', color='grey')

        _groups = self.gitlab.groups.list(get_all=True, search=group_name)

        print_info(_groups, color='grey')
        assert len(_groups) >= 1, f'Could not find group_name={group_name}'
        group = _groups[0]  # type: ignore

        project = {i.name: i for i in group.projects.list(all=True)}[project_name]

        project = typing.cast(gitlab.v4.objects.GroupProject, project)
        print_info(f'Got private project: <{project.name}>', color='grey')

        return project

    def get_public_project(
            self,
            private_group_name: str,
            public_repo_name: str,
    ) -> gitlab.v4.objects.GroupProject:
        return self.get_project_from_group(private_group_name, public_repo_name)

    def get_projects_in_group(
            self,
            group_name: str,
    ) -> list[gitlab.v4.objects.GroupProject]:
        print_info(f'Get projects in group_name={group_name}', color='grey')

        _groups = self.gitlab.groups.list(get_all=True, search=group_name)

        assert len(_groups) >= 1, f'Could not find group_name={group_name}'
        group = _groups[0]  # type: ignore

        print_info(f'Got group: <{group.name}>', color='grey')

        projects = group.projects.list(all=True)

        projects = typing.cast(list[gitlab.v4.objects.GroupProject], projects)
        print_info(f'Got {len(projects)} projects', color='grey')

        return projects

    def get_group_members(
            self,
            group_name: str,
    ) -> list[gitlab.v4.objects.GroupMember]:
        print_info(f'Get members in group_name={group_name}', color='grey')

        _groups = self.gitlab.groups.list(get_all=True, search=group_name)

        print_info(_groups, color='grey')
        assert len(_groups) >= 1, f'Could not find group_name={group_name}'
        group = _groups[0]  # type: ignore

        print_info(f'Got group: <{group.name}>', color='grey')

        members = group.members_all.list(all=True, get_all=True)

        members = typing.cast(list[gitlab.v4.objects.GroupMember], members)
        print_info(f'Got {len(members)} members', color='grey')

        return members

    def get_project_members(
            self,
            project_name: str,
    ) -> list[gitlab.v4.objects.ProjectMember]:
        print_info(f'Get members in project_name={project_name}', color='grey')

        _projects = self.gitlab.projects.list(get_all=True, search=project_name)

        assert len(_projects) >= 1, f'Could not find project_name={project_name}'
        project = _projects[0]  # type: ignore

        print_info(f'Got project: <{project.name}>', color='grey')

        members = project.members_all.list(all=True, get_all=True)

        members = typing.cast(list[gitlab.v4.objects.ProjectMember], members)
        print_info(f'Got {len(members)} members', color='grey')

        return members

    def get_user_by_username(
            self,
            username: str,
    ) -> gitlab.v4.objects.User:
        print_info(f'Get user with username={username}', color='grey')

        _users = self.gitlab.users.list(get_all=True, search=username)
        assert len(_users) > 0, f'Could not find username={username}'

        if len(_users) > 1:
            print_info(
                f'Got multiple users: <{[(user.username, user.name) for user in _users]}>',
                color='grey'
            )
            _username_to_user = {user.username: user for user in _users}
            assert username in _username_to_user, f'Could not find username={username}'
            user = _username_to_user[username]
        else:
            user = _users[0]  # type: ignore

        user = typing.cast(gitlab.v4.objects.User, user)
        print_info(f'Got user: <{user.name}>', color='grey')

        return user

    def get_all_tutors(
            self,
            private_group_name: str,
    ) -> list[gitlab.v4.objects.GroupMember]:
        return self.get_group_members(private_group_name)

    def get_students_projects(
            self,
            students_group_name: str,
    ) -> list[gitlab.v4.objects.GroupProject]:
        return self.get_projects_in_group(students_group_name)

    def get_student_file_link(
            self,
            gitlab_url: str,
            default_branch: str,
            students_group_name: str,
            username: str,
            path: str | Path,
    ) -> str:
        return f'{gitlab_url}/{students_group_name}/{username}/-/blob/{default_branch}/{path}'

    def get_current_user(
            self,
    ) -> gitlab.v4.objects.CurrentUser:
        self.gitlab.auth()
        current_user = self.gitlab.user

        current_user = typing.cast(gitlab.v4.objects.CurrentUser, current_user)

        return current_user

    def get_group(
            self,
            name: str,
    ) -> gitlab.v4.objects.Group:
        print_info(f'Get group name={name}', color='grey')

        _groups = self.gitlab.groups.list(get_all=True, search=name)

        assert len(_groups) >= 1, f'Could not find group name={name}'
        group = _groups[0]  # type: ignore

        group = typing.cast(gitlab.v4.objects.Group, group)

        return group



from __future__ import annotations

import copy
from typing import Any

import pytest

from checker.exceptions import BadConfig
from checker.pipeline import ParametersResolver


class TestParametersResolver:
    @pytest.mark.parametrize(
        "template, context, expected",
        [
            ("${{ a }}", {"a": 2}, 2),
            pytest.param("${{ b }}", {"b": "2"}, "2", marks=pytest.mark.xfail()),  # TODO: check why returned as int
            ("${{ c }}", {"c": [1, 2, "4"]}, [1, 2, "4"]),
            ("    ${{ d }}", {"d": 2}, 2),
            ("${{ e }}   ", {"e": 2}, 2),
            ("${{ f }} some string", {"f": 2}, "2 some string"),
            ("${{ g }} + ${{ g }}", {"g": 2}, "2 + 2"),
            ("${{ h }}", {"h": 2.1}, 2.1),
            ("${{ i }}", {"i": 2.0}, 2.0),
        ],
    )
    def test_keep_native_type(self, template: str, context: dict[str, Any], expected: Any) -> None:
        resolver = ParametersResolver()
        assert resolver.resolve(template, context) == expected

    @pytest.mark.parametrize(
        "template, context, expected",
        [
            ("${{ a }}", {"a": 2}, 2),
            ("Hello, ${{ name }}!", {"name": "World"}, "Hello, World!"),
            ("${{ a }} + ${{ b }} = ${{ a + b }}", {"a": 2, "b": 3}, "2 + 3 = 5"),
            ("${{ a }}", {"a": 2, "b": 3}, 2),
        ],
    )
    def test_string_input(self, template: str, context: dict[str, Any], expected: Any) -> None:
        resolver = ParametersResolver()
        assert resolver.resolve(template, context) == expected

    @pytest.mark.parametrize(
        "template, context, expected",
        [
            (["${{ item }}", "${{ item }}2"], {"item": "test"}, ["test", "test2"]),
            (["${{ a }}", "${{ b }}"], {"a": 1, "b": 2}, [1, 2]),
            (
                ["${{ a }}", ["${{ b }}", "${{ c }}"]],
                {"a": 1, "b": 2, "c": 3},
                [1, [2, 3]],
            ),
        ],
    )
    def test_list_input(self, template: list[Any], context: dict[str, Any], expected: list[Any]) -> None:
        resolver = ParametersResolver()
        assert resolver.resolve(template, context) == expected

    @pytest.mark.parametrize(
        "template, context, expected",
        [
            (
                {"key1": "${{ a }}", "key2": "${{ b }}"},
                {"a": "x", "b": "y"},
                {"key1": "x", "key2": "y"},
            ),
            (
                {"name": "Hello, ${{ name }}!"},
                {"name": "Alice"},
                {"name": "Hello, Alice!"},
            ),
            (
                {"key1": "${{ a }}", "key2": {"key3": "${{ b }}"}},
                {"a": 1, "b": 2},
                {"key1": 1, "key2": {"key3": 2}},
            ),
        ],
    )
    def test_dict_input(
        self,
        template: dict[str, Any],
        context: dict[str, Any],
        expected: dict[str, Any],
    ) -> None:
        resolver = ParametersResolver()
        assert resolver.resolve(template, context) == expected

    @pytest.mark.parametrize(
        "template, context",
        [
            (1, {}),
            (1, {"a": 1}),
            (1.0, {"a": 1}),
            ("some string", {"a": 1}),
            ("a", {"a": 1}),
            ("{a}", {"a": 1}),
            ({}, {"a": 1}),
            ([None, {1, 2, 3}, ["a", "b"]], {"a": 1}),
        ],
    )
    def test_non_template(self, template: Any, context: dict[str, Any]) -> None:
        resolver = ParametersResolver()
        template_copy = copy.deepcopy(template)
        assert resolver.resolve(template, context) == template_copy

    @pytest.mark.parametrize(
        "template, context",
        [
            ("${{ invalid_syntax", {"invalid_syntax": 2}),
            pytest.param(
                "${{ valid_var.invalid_field }}",
                {"valid_var": {"valid_field": 1}},
                marks=pytest.mark.xfail(),
            ),
            pytest.param("${{ not_existing }} ${{ a }}", {"a": 2}, marks=pytest.mark.xfail()),
            pytest.param("${{ not_existing }}", {"a": 2}, marks=pytest.mark.xfail()),
            pytest.param("invalid_syntax }}", {"invalid_syntax": 2}, marks=pytest.mark.xfail()),
        ],
    )
    def test_invalid_template(self, template: Any, context: dict[str, Any]) -> None:
        resolver = ParametersResolver()
        with pytest.raises(BadConfig):
            a = resolver.resolve(template, context)
            print(a)
